##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
UDS Auxiliary plugin
********************

:module: uds_auxiliary

:synopsis: implementation of existing UdsAuxiliary for Robot
    framework usage.

.. currentmodule:: uds_auxiliary

"""

from typing import List, Optional, Union

from robot.api import logger
from robot.api.deco import keyword, library
from uds import IsoServices

from pykiso.lib.robot_framework.aux_interface import RobotAuxInterface

from ..auxiliaries.udsaux.common.uds_response import UdsResponse
from ..auxiliaries.udsaux.uds_auxiliary import UdsAuxiliary as UdsAux


@library(version="0.1.0")
class UdsAuxiliary(RobotAuxInterface):
    """Robot framework plugin for UdsAuxiliary."""

    ROBOT_LIBRARY_SCOPE = "SUITE"

    def __init__(self):
        """Initialize attributes."""
        super().__init__(aux_type=UdsAux)

    @keyword(name="Send uds raw")
    def send_uds_raw(
        self,
        msg_to_send: bytes,
        aux_alias: str,
        timeout_in_s: float = 6,
        response_required: bool = True,
    ) -> Union[list, bool]:
        """Send a UDS diagnostic request to the target ECU.

        :param msg_to_send: can uds raw bytes to be sent
        :param aux_alias: auxiliary's alias
        :param timeout_in_s: maximum time used to wait for a response.
        :param response_required: Wait for a response if True

        :return: the raw uds response's, or True if a response is
            not expected and the command is properly sent otherwise
            False
        """
        aux = self._get_aux(aux_alias)
        msg = aux.send_uds_raw(msg_to_send, timeout_in_s, response_required)
        return list(msg) if not isinstance(msg, bool) else msg

    @keyword(name="Send uds config")
    def send_uds_config(self, msg_to_send: dict, aux_alias: str, timeout_in_s: float = 6) -> Union[dict, bool]:
        """Send UDS config to the target ECU.

        :param msg_to_send: uds config to be sent
        :param aux_alias: auxiliary's alias
        :param timeout_in_s: maximum time used to wait for a response

        :return: a dict containing the uds response, or True if a
            response is not expected and the command is properly sent
            otherwise False
        """
        msg_to_send["service"] = self.get_service_id(msg_to_send["service"])
        aux = self._get_aux(aux_alias)
        return aux.send_uds_config(msg_to_send, timeout_in_s)

    @keyword(name="Check raw response positive")
    def check_raw_response_positive(self, resp: UdsResponse, aux_alias: str) -> Optional[bool]:
        """Check if the response is positive, raise an error if not

        :param resp: raw response of uds request

        :raise UnexpectedResponseError: raised when the answer is not the expected one

        :return: True if response is positive
        """
        aux = self._get_aux(aux_alias)
        return aux.check_raw_response_positive(resp)

    @keyword(name="Check raw response negative")
    def check_raw_response_negative(self, resp: UdsResponse, aux_alias: str) -> Optional[bool]:
        """Check if the response is negative, raise an error if not

        :param resp: raw response of uds request

        :raise UnexpectedResponseError: raised when the answer is not the expected one

        :return: True if response is negative
        """
        aux = self._get_aux(aux_alias)
        return aux.check_raw_response_negative(resp)

    @keyword(name="Soft reset")
    def soft_reset(self, aux_alias: str) -> Union[dict, List[int]]:
        """Perform soft reset of the component, equivalent to a restart of application

        :param aux_alias: auxiliary's alias
        """
        aux = self._get_aux(aux_alias)
        return aux.soft_reset()

    @keyword(name="Hard reset")
    def hard_reset(self, aux_alias: str) -> Union[dict, List[int]]:
        """Allow power reset of the component

        :param aux_alias: auxiliary's alias
        """
        aux = self._get_aux(aux_alias)
        return aux.hard_reset()

    @keyword(name="Force ecu reset")
    def force_ecu_reset(self, aux_alias: str) -> List[int]:
        """Allow power reset of the component

        :param aux_alias: auxiliary's alias
        """
        aux = self._get_aux(aux_alias)
        return aux.force_ecu_reset()

    @keyword(name="Read data")
    def read_data(self, parameter: str, aux_alias: str) -> Union[dict, bool]:
        """UDS config command that allow data reading

        :param parameter: data to be read
        :param aux_alias: auxiliary's alias

        :return: a dict with uds config response
        """
        aux = self._get_aux(aux_alias)
        return aux.read_data(parameter)

    @keyword(name="Write data")
    def write_data(self, parameter: str, value: Union[List[bytes], bytes], aux_alias: str) -> Union[dict, bool]:
        """UDS config command that allow data reading

        :param parameter: data to be read
        :param value: new content of the data
        :param aux_alias: auxiliary's alias

        :return: a dict with uds config response
        """
        aux = self._get_aux(aux_alias)
        return aux.write_data(parameter, value)

    @keyword(name="Start tester present with ${period} seconds ${aux_alias} ")
    def start_tester_present_sender(self, aux_alias, period: int = 4) -> None:
        """Start to continuously sends tester present messages via UDS

        :param period: period in seconds to use for the cyclic sending of tester present
        :param aux_alias: auxiliary's alias
        """
        aux = self._get_aux(aux_alias)
        aux.start_tester_present_sender(period)

    @keyword(name="Stop tester present")
    def stop_tester_present_sender(self, aux_alias) -> None:
        """Stop to continuously sends tester present messages via UDS"""
        aux = self._get_aux(aux_alias)
        aux.stop_tester_present_sender()

    @staticmethod
    def get_service_id(service_name: str) -> int:
        """Return the uds service id.

        :param service_name: service's name (EcuReset,
            ReadDataByIdentifier ...)

        :return: corresponding service identification number
        """
        try:
            service_id = IsoServices[service_name].value
        except KeyError:
            logger.error(f"{service_name} doesn't exist", html=True)
            raise KeyError(f"{service_name} doesn't exist")

        return service_id



##########################################################################
# Copyright (c) 2024-2024 Accenture GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
REST API Endpoint
*****************

:module: rest_server

:synopsis: Expose pykiso setup as a rest server

.. currentmodule:: rest_server
"""
from typing import Callable

import uvicorn
from fastapi import FastAPI

from pykiso.auxiliary import AuxiliaryInterface
from pykiso.endpoint import Endpoint


class RestServer(Endpoint):
    """Class to expose pykiso setup as a REST server."""

    def __init__(self, host: str, port: int):
        """Constructor

        :param host: host ip address
        :param port: host port

        """
        self.app = FastAPI(title="Pykiso REST Server")
        self.host = host
        self.port = port
        self.list_of_auxiliaries = []
        super().__init__()

    def expose_one_auxiliary(self, alias: str, auxiliary: AuxiliaryInterface):
        """Implements the parent method to expose one auxiliary.

        :param alias: Given auxiliary name from the yaml file
        :param auxiliary: The auxiliary instance associated with the alias

        """
        # Add to list of auxiliaries
        self.list_of_auxiliaries.append(alias)
        for method in dir(auxiliary):
            if (
                method.startswith("_")
                or not isinstance(getattr(auxiliary, method), Callable)
                or method == "get_instance"
                or "proxy" in method
                or "run_command" == method
                or method == "collect_messages"
            ):
                continue

            def bindFunction(name):
                return self.app.post("/" + alias + "/" + method)(getattr(auxiliary, name))

            bindFunction(method)

    def wait_for_termination(self):
        """Start of the REST server"""

        @self.app.get("/")
        async def root():
            return {"message": "Welcome to Pykiso", "auxiliaries": self.list_of_auxiliaries}

        uvicorn.run(
            self.app,
            host=self.host,
            port=self.port,
        )



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Proxy Channel
*************

:module: cc_proxy

:synopsis: CChannel implementation for multi-auxiliary usage.

CCProxy channel was created, in order to enable the connection of
multiple auxiliaries on one and only one CChannel. This CChannel
has to be used with a so called proxy auxiliary.

.. currentmodule:: cc_proxy

"""
from __future__ import annotations

import logging
import queue
import threading
from typing import TYPE_CHECKING, Any, Callable

from pykiso.connector import CChannel

if TYPE_CHECKING:
    from pykiso.lib.auxiliaries.proxy_auxiliary import ProxyAuxiliary
    from pykiso.types import ProxyReturn


log = logging.getLogger(__name__)


class CCProxy(CChannel):
    """Proxy CChannel to bind multiple auxiliaries to a single 'physical' CChannel."""

    # keep an uninitialized proxy variable at class-level to fulfill the existence conditions.
    # when a CCProxy instance will then set it, it will only be set at instance level but stay
    # uninitialized at class-level
    _proxy: ProxyAuxiliary = None
    _physical_channel: CChannel = None

    def __init__(self, **kwargs):
        """Initialize attributes."""
        super().__init__(**kwargs)
        self.queue_out = None
        self.timeout = 1
        self._lock = threading.Lock()
        self._tx_callback = None

    def _bind_channel_info(self, proxy_aux: ProxyAuxiliary):
        """Bind a :py:class:`~pykiso.lib.auxiliaries.proxy_auxiliary.ProxyAuxiliary`
        instance that is instantiated in order to handle the connection of
        multiple auxiliaries to a single communication channel.

        This allows to access the real communication channel's attributes
        and hides the underlying proxy setup.

        :param proxy_aux: the proxy auxiliary instance that is holding the
            real communication channel.
        """
        self._proxy = proxy_aux
        self._physical_channel = proxy_aux.channel

    def __getattr__(self, name: str) -> Any:
        """Implement getattr to retrieve attributes from the real channel attached
        to the underlying :py:class:`~pykiso.lib.auxiliaries.proxy_auxiliary.ProxyAuxiliary`.

        :param name: name of the attribute to get.
        :raises AttributeError: if the attribute is not part of the real
            channel instance or if the real channel hasn't been bound to
            this proxy channel yet.
        :return: the found attribute value.
        """
        if self._physical_channel is not None:
            with self._proxy.lock:
                return getattr(self._physical_channel, name)
        raise AttributeError(f"{self.__class__.__name__} object has no attribute '{name}'")

    def detach_tx_callback(self) -> None:
        """Detach the current callback."""
        with self._lock:
            log.internal_warning("reset current attached transmit callback!")
            self._tx_callback = None

    def attach_tx_callback(self, func: Callable) -> None:
        """Attach to a callback to the _cc_send method.

        :param func: function to call when _cc_send is called
        """
        with self._lock:
            log.internal_debug(f"attached function {func.__name__}")
            if self._tx_callback is not None:
                log.internal_warning(
                    f"function {func.__name__} will replace current transmit callback {self._tx_callback.__name__}"
                )
            self._tx_callback = func

    def open(self) -> None:
        super().open()
        if self._proxy is not None:
            # will start the proxy_auxiliary if this is the first CCProxy to be opened
            self._proxy._open_connections += 1

    def close(self) -> None:
        super().close()
        if self._proxy is not None:
            # will stop the proxy_auxiliary if this is the last CCProxy to be closed
            self._proxy._open_connections -= 1

    def _cc_open(self) -> None:
        """Open proxy channel."""
        log.internal_info("Open proxy channel")
        self.queue_out = queue.Queue()

    def _cc_close(self) -> None:
        """Close proxy channel."""
        log.internal_debug("Close proxy channel")
        self.queue_out = None

    def _cc_send(self, *args: Any, **kwargs: Any) -> None:
        """Call the attached ProxyAuxiliary's transmission callback
        with the provided arguments.

        :param args: positionnal arguments to pass to the callback
        :param kwargs: named arguments to pass to the callback
        """
        log.internal_debug(f"put at proxy level: {args} {kwargs}")
        if self._tx_callback is not None:
            # call the attached ProxyAuxiliary's run_command method
            self._tx_callback(self, *args, **kwargs)

    def _cc_receive(self, timeout: float = 0.1) -> ProxyReturn:
        """Depopulate the queue out of the proxy connector.

        :param timeout: not used

        :return: bytes and source when it exist. if queue timeout
            is reached return None
        """
        try:
            return_response = self.queue_out.get(True, self.timeout)
            log.internal_debug(f"received at proxy level : {return_response}")
            return return_response
        except queue.Empty:
            return {"msg": None}



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Communication Channel Via Serial
********************************

:module: cc_serial

:synopsis: Serial communication channel

.. currentmodule:: cc_serial

"""

import logging
import sys
import time
from enum import Enum, IntEnum
from typing import ByteString, Dict, List, Optional, Union

try:
    import serial
    import serial.tools.list_ports
except ImportError as e:
    raise ImportError(f"{e.name} dependency missing, consider installing pykiso with 'pip install pykiso[serial]'")

log = logging.getLogger(__name__)

from pykiso import Message, connector

MessageType = Union[Message, bytes]


class ByteSize(IntEnum):
    FIVE_BITS = serial.FIVEBITS
    SIX_BITS = serial.SIXBITS
    SEVEN_BITS = serial.SEVENBITS
    EIGHT_BITS = serial.EIGHTBITS


class Parity(str, Enum):
    PARITY_NONE = serial.PARITY_NONE
    PARITY_EVEN = serial.PARITY_EVEN
    PARITY_ODD = serial.PARITY_ODD
    PARITY_MARK = serial.PARITY_MARK
    PARITY_SPACE = serial.PARITY_SPACE


class StopBits(IntEnum):
    STOPBITS_ONE = serial.STOPBITS_ONE
    STOPBITS_ONE_POINT_FIVE = serial.STOPBITS_ONE_POINT_FIVE
    STOPBITS_TWO = serial.STOPBITS_TWO


class CCSerial(connector.CChannel):
    """Connector for serial devices"""

    def __init__(
        self,
        port: Optional[str] = None,
        serial_number: Optional[str] = None,
        vid: Optional[int] = None,
        pid: Optional[int] = None,
        baudrate: int = 9600,
        bytesize: ByteSize = ByteSize.EIGHT_BITS,
        parity: Parity = Parity.PARITY_NONE,
        stopbits: StopBits = StopBits.STOPBITS_ONE,
        timeout: Optional[float] = None,
        xonxoff: bool = False,
        rtscts: bool = False,
        write_timeout: Optional[float] = None,
        dsrdtr: bool = False,
        inter_byte_timeout: Optional[float] = None,
        exclusive: Optional[bool] = None,
        **kwargs,
    ):
        """Init Serial settings

        :param port: Device name (e.g. "COM1" for Windows or "/dev/ttyACM0" for Linux)
        :param serial_number: serial number
        :param vid: vendor id
        :param pid: product id
        :param baudrate: Baud rate such as 9600 or 115200 etc, defaults to 9600
        :param bytesize: Number of data bits. Possible values:
          FIVEBITS, SIXBITS, SEVENBITS, EIGHTBITS, defaults to EIGHTBITS
        :param parity: Enable parity checking. Possible values:
          PARITY_NONE, PARITY_EVEN, PARITY_ODD PARITY_MARK, PARITY_SPACE,
          defaults to PARITY_NONE
        :param stopbits:  Number of stop bits. Possible values:
          STOPBITS_ONE, STOPBITS_ONE_POINT_FIVE, STOPBITS_TWO,
          defaults to STOPBITS_ONE
        :param timeout: Set a read timeout value in seconds, defaults to None
        :param xonxoff: Enable software flow control, defaults to False
        :param rtscts: Enable hardware (RTS/CTS) flow control, defaults to False
        :param write_timeout: Set a write timeout value in seconds, defaults to None
        :param dsrdtr: Enable hardware (DSR/DTR) flow control, defaults to False
        :param inter_byte_timeout:  Inter-character timeout, None to disable,
          defaults to None
        :param exclusive: Set exclusive access mode (POSIX only).
          A port cannot be opened in exclusive access mode if it is already open
          in exclusive access mode., defaults to None
        """

        # Initialize the super class
        super().__init__(**kwargs)
        # Initialize the serial connection. Port None prevents automatic opening
        self.serial = serial.Serial(
            port=None,
            baudrate=baudrate,
            bytesize=bytesize,
            parity=parity,
            stopbits=stopbits,
            timeout=timeout,
            xonxoff=xonxoff,
            rtscts=rtscts,
            write_timeout=write_timeout,
            dsrdtr=dsrdtr,
            inter_byte_timeout=inter_byte_timeout,
            exclusive=exclusive,
        )

        self.current_write_timeout = write_timeout
        self.serial.port = self._get_port(port=port, vid=vid, pid=pid, serial_number=serial_number)

    @staticmethod
    def _find_device(vid: int, pid: int, serial_number: str) -> str:
        """Return the device which matches the given pid/vid or serial_number.

        :param vid: vendor id
        :param pid: product id
        :param serial_number: serial number

        :return: com port for the found device. I.e. "COM4" or "/dev/tty1"
        """

        attached_com_devices = serial.tools.list_ports.comports()
        found_devices = []
        if pid and vid:
            found_devices.extend([port for port in attached_com_devices if port.pid == pid and port.vid == vid])

        if serial_number:
            found_devices.extend([port for port in attached_com_devices if port.serial_number == serial_number])

        if not found_devices:
            raise ConnectionError(
                f"Failed to detect connected USB device with IDs {vid:04X}:{pid:04x} or serial_number {serial_number}."
            )

        found_devices = [port.name if not sys.platform.startswith("win") else port.device for port in found_devices]
        if len(found_devices) > 1:
            log.internal_warning(
                f"Found multiple devices, {found_devices}, with matching IDs {vid:04X}:{pid:04X} or serial_number {serial_number}. Select first device {found_devices[0]}."
            )
        return found_devices[0]

    def _get_port(self, port: str, vid: int, pid: int, serial_number: str) -> str:
        """Returns com port depending on the given port argument.
        If port is None, the device will be searched for,
        using the pid/vid and serial_number, else the port argument will be returned.

        :param port: port  I.e. "COM4" or "/dev/tty1"
        :param vid: vendor id
        :param pid: product id

        :return: com ports of given or found device. I.e. "COM4" or "/dev/tty1"
        """
        if port is None:
            return CCSerial._find_device(vid=vid, pid=pid, serial_number=serial_number)
        else:
            return port

    def _cc_open(self) -> None:
        """Open serial port"""
        self.serial.open()

    def _cc_close(self) -> None:
        """Close serial port"""
        self.serial.close()

    def _cc_send(self, msg: Union[ByteString, str], timeout: float = None, **kwargs) -> None:
        """Sends data to the serial port

        :param msg: data to send
        :param timeout: write timeout in seconds. None sets it to blocking,
          defaults to None
        :param kwargs: not used
        :raises: SerialTimeoutException - In case a write timeout is configured
          for the port and the time is exceeded.
        """
        if timeout != self.current_write_timeout:
            self.current_write_timeout = timeout
            self.serial.write_timeout = timeout

        # Some auxiliaries pass msg as a string
        if isinstance(msg, str):
            msg = msg.encode()

        self.serial.write(msg)

    def _cc_receive(self, timeout=0.00001) -> Dict[str, Optional[bytes]]:
        """Read bytes from the serial port.
        Try to read one byte in blocking mode. After blocking read check
        remaining bytes and read them without a blocking call.

        :param timeout: timeout in seconds, 0 for non blocking read, defaults to 0.00001
        :raises NotImplementedError: if raw is to True
        :return: received bytes
        """

        self.serial.timeout = timeout

        received = self.serial.read()

        # We have waited already so set the timeout to non blocking
        self.serial.timeout = 0

        # Check how many bytes are still left and read them at once
        in_waiting = self.serial.in_waiting
        if in_waiting > 0:
            received += self.serial.read(in_waiting)

        return {"msg": received}



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
uds_response
************

:module: uds_response

:synopsis: This module contains an enum listing the
    UDS negative response codes and a wrapper around
    raw UDS responses.

.. currentmodule:: uds_response
"""
import logging
from collections import UserList
from enum import IntEnum
from typing import List

log = logging.getLogger(__name__)


class UdsResponse(UserList):
    """List wrapper adding attributes relative to a UDS response."""

    NEGATIVE_RESPONSE_SID = 0x7F

    def __init__(
        self,
        response_data: List[int],
        resp_time: float = None,
        pending_resp_times: List[float] = None,
    ) -> None:
        """Initialize attributes.

        :param response_data: the original response data.
        :param resp_time: time to get the response in seconds
        :param pending_resp_times: list of the times between each response pending, in seconds
        """
        super().__init__(response_data)
        self.is_negative = False
        self.nrc = None
        if self.data and self.data[0] == self.NEGATIVE_RESPONSE_SID:
            self.is_negative = True
            self.nrc = NegativeResponseCode(self.data[2])
        self.resp_time = resp_time
        self.pending_resp_times = pending_resp_times

    def __repr__(self):
        if self.data:
            if self.is_negative:
                return f"NegativeUdsResponse({bytes(self.data).hex(sep=' ')}, nrc={self.nrc.name})"
            return f"{self.__class__.__name__}({bytes(self.data).hex(sep=' ')})"
        else:
            return f"{self.__class__.__name__} no data."


class NegativeResponseCode(IntEnum):
    """NRC code response"""

    GENERAL_REJECT = 0x10
    SERVICE_NOT_SUPPORTED = 0x11
    SUB_FUNCTION_NOT_SUPPORTED = 0x12
    INVALID_FORMAT = 0x13
    RESPONSE_TOO_LONG = 0x14
    BUSY_REPEAT_REQUEST = 0x21
    CONDITIONS_NOT_CORRECT = 0x22
    REQUEST_SEQUENCE_ERROR = 0x24
    NO_RESPONSE_FROM_SUB_NET_COMPONENT = 0x25
    FAILURE_PREVENTS_EXECUTION = 0x26
    REQUEST_OUT_OF_RANGE = 0x31
    SECURITY_ACCESS_DENIED = 0x33
    AUTHENTICATION_REQUIRED = 0x34
    INVALID_KEY = 0x35
    EXCEED_NUMBER_ATTEMPTS = 0x36
    REQUIRED_TIME_DELAY_NOT_EXPIRED = 0x37
    SECURE_DATA_TRANSMISSION_REQUIRED = 0x38
    SECURE_DATA_TRANSMISSION_NOT_ALLOWED = 0x39
    SECURE_DATA_VERIFICATION_FAILED = 0x3A
    CERTIFICATE_VERIFICATION_FAILED_INVALID_TIME_PERIODE = 0x50
    CERTIFICATE_VERIFICATION_FAILED_INVALID_SIGNATURE = 0x51
    CERTIFICATE_VERIFICATION_FAILED_INVALID_CHAIN_OF_TRUST = 0x52
    CERTIFICATE_VERIFICATION_FAILED_INVALID_TYPE = 0x53
    CERTIFICATE_VERIFICATION_FAILED_INVALID_FORMAT = 0x54
    CERTIFICATE_VERIFICATION_FAILED_INVALID_CONTENT = 0x55
    CERTIFICATE_VERIFICATION_FAILED_INVALID_SCOPE = 0x56
    CERTIFICATE_VERIFICATION_FAILED_INVALID_CERTIFICATE_REVOKED = 0x57
    OWNERSHIP_VERIFICATION_FAILED = 0x58
    CHALLENGE_CALCULATION_FAILED = 0x59
    SETTING_ACCESS_RIGHT_FAILED = 0x5A
    SESSION_KEY_CREATION_DERIVATION_FAILED = 0x5B
    CONFIGURATION_DATA_USAGE_FAILED = 0x5C
    DE_AUTHENTICATION_FAILED = 0x5D
    UPLOAD_DOWNLOAD_NOT_ACCEPTED = 0x70
    TRANSFER_DATA_SUSPENDED = 0x71
    GENERAL_PROGRAMMING_FAILURE = 0x72
    WRONG_BLOCK_SEQUENCE_COUNTER = 0x73
    REQUEST_CORRECTLY_RECEIVED_RESPONSE_PENDING = 0x78
    SUBFUNCTION_NOT_SUPPORTED_IN_ACTIVE_SESSION = 0x7E
    SERVICE_NOT_SUPPORTED_IN_ACTIVE_SESSION = 0x7F
    RPM_TOO_HIGH = 0x81
    RPM_TOO_LOW = 0x82
    ENGINE_IS_RUNNING = 0x83
    ENGINE_IS_NOT_RUNNING = 0x84
    ENGINE_TIME_TOO_LOW = 0x85
    TEMPERATURE_TOO_HIGH = 0x86
    TEMPERATURE_TOO_LOW = 0x87
    VEHICULE_SPEED_TOO_HIGH = 0x88
    VEHICULE_SPEED_TOO_LOW = 0x89
    THROTTLE_PEDAL_TOO_LOW = 0x8A
    THROTTLE_PEDAL_TOO_HIGH = 0x8B
    TRANSMISSION_RANGE_NOT_IN_NEUTRAL = 0x8C
    TRANSMISSION_RANGE_NOT_IN_GEAR = 0x8D
    BREAK_SWITCH_NOT_CLOSED = 0x8F
    SHIFTER_LEVER_NOT_IN_PARK = 0x90
    TORQUE_CONVERTER_CLUTCH_LOCKED = 0x91
    VOLTAGE_TOO_HIGH = 0x92
    VOLTAGE_TOO_LOW = 0x93
    RESSOURCE_TEMPORARLY_NOT_AVAILABLE = 0x94

    @classmethod
    def _missing_(cls, value: int) -> int:
        """Instead of raising a ValueError on a missing member,
        create a NRC UNKNOWN and return it.

        :param value: missing NRC value.
        :return: a new NRC member UNKNOWN with the missing value.
        """
        log.internal_warning(f"Unknown NRC: {value}")
        new_member = int.__new__(cls, value)
        new_member._name_ = "UNKNOWN"
        new_member._value_ = value
        return new_member



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
uds_utils
*********

:module: uds_utils

:synopsis: This module contains service names functions enums.

.. currentmodule:: util_utils
"""

import logging
from enum import IntEnum

log = logging.getLogger(__name__)


SERVICE_ID_TO_NAME = {
    0x10: "diagnosticSessionControl",
    0x11: "ecuReset",
    0x14: "clearDTC",
    0x19: "readDTC",
    0x22: "readDataByIdentifier",
    0x27: "securityAccess",
    0x2E: "writeDataByIdentifier",
    0x2F: "inputOutputControl",
    0x31: "routineControl",
    0x34: "requestDownload",
    0x35: "requestUpload",
    0x36: "transferData",
    0x37: "transferExit",
    0x3E: "testerPresent",
}


def get_uds_service(service_id: IntEnum) -> str or None:
    """Get uds service name function that is binded to uds instance

    :param service_id: uds service identifier

    :return: uds service function name
    """
    name = SERVICE_ID_TO_NAME.get(service_id)
    if name is None:
        log.error("UDS service ID not found")

    return name



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################


"""
TestRail tool command line interface
************************************

:module: cli

:synopsis: CLI entry-point for TestRail helper tool

.. currentmodule:: cli

"""
import getpass
import sys

import click

from .console import print_cases, print_milestones, print_projects, print_results, print_runs, print_suites
from .extraction import JunitReport
from .testrail import (
    add_case_results,
    create_new_run,
    enumerate_all_cases,
    enumerate_all_milestones,
    enumerate_all_projects,
    enumerate_all_runs,
    enumerate_all_suites,
    extract_test_results,
    search_for_milestone,
    search_for_suite,
)


@click.group()
@click.option(
    "--user",
    help="TestRail user id",
    required=True,
    default=None,
    hide_input=True,
)
@click.option(
    "--password",
    help="Valid TestRail API key (if not given ask at command prompt level)",
    required=False,
    default=None,
    hide_input=True,
)
@click.option(
    "--url",
    help="URL of TestRail server",
    required=True,
)
@click.pass_context
def cli_testrail(ctx: dict, user: str, password: str, url: str) -> None:
    """TestRail interaction tool."""
    ctx.ensure_object(dict)
    ctx.obj["USER"] = user or input("Enter Client ID TestRail and Press enter:")
    ctx.obj["PASSWORD"] = password or getpass.getpass("Enter your password and Press ENTER:")
    ctx.obj["URL"] = url


@cli_testrail.command("projects")
@click.pass_context
def cli_projects(ctx) -> None:
    """Returns the list of available projects."""
    projects = enumerate_all_projects(
        base_url=ctx.obj["URL"],
        user=ctx.obj["USER"],
        password=ctx.obj["PASSWORD"],
    )
    print_projects(projects)


@cli_testrail.command("suites")
@click.option(
    "-p",
    "--project",
    help="TestRail Project's name",
    required=True,
    default=None,
    type=click.STRING,
)
@click.pass_context
def cli_suites(ctx, project: str) -> None:
    """Returns a list of all the test suites contained in a given project."""
    _, suites = enumerate_all_suites(
        base_url=ctx.obj["URL"],
        user=ctx.obj["USER"],
        password=ctx.obj["PASSWORD"],
        project_name=project,
    )
    print_suites(project, suites)


@cli_testrail.command("cases")
@click.option(
    "-p",
    "--project",
    help="TestRail Project's name",
    required=True,
    default=None,
    type=click.STRING,
)
@click.option(
    "--custom-field",
    help="TestRail's case custom field use to store the requirement id",
    required=False,
    default="id",
    type=click.STRING,
)
@click.pass_context
def cli_cases(
    ctx,
    project: str,
    custom_field: str,
) -> None:
    """Returns a list of all cases contained in a project."""
    _, cases = enumerate_all_cases(
        base_url=ctx.obj["URL"],
        user=ctx.obj["USER"],
        password=ctx.obj["PASSWORD"],
        project_name=project,
        custom_field=custom_field,
    )
    print_cases(project, cases)


@cli_testrail.command("runs")
@click.option(
    "-p",
    "--project",
    help="TestRail's project name",
    required=True,
    default=None,
    type=click.STRING,
)
@click.pass_context
def cli_runs(ctx, project: str) -> None:
    """Returns a list of all runs contained in a project."""
    _, runs = enumerate_all_runs(
        base_url=ctx.obj["URL"],
        user=ctx.obj["USER"],
        password=ctx.obj["PASSWORD"],
        project_name=project,
    )
    print_runs(project, runs)


@cli_testrail.command("milestones")
@click.option(
    "-p",
    "--project",
    help="TestRail's project name",
    required=True,
    default=None,
    type=click.STRING,
)
@click.pass_context
def cli_milestones(ctx, project: str) -> None:
    """Returns the list of all milestones contained in a project."""
    _, milestones = enumerate_all_milestones(
        base_url=ctx.obj["URL"],
        user=ctx.obj["USER"],
        password=ctx.obj["PASSWORD"],
        project_name=project,
    )
    print_milestones(project, milestones)


@cli_testrail.command("upload")
@click.option(
    "-n",
    "--run-name",
    help="How to name the created run on TestRail",
    required=True,
    default=None,
    type=click.STRING,
)
@click.option(
    "-p",
    "--project",
    help="TestRail's project name",
    required=True,
    default=None,
    type=click.STRING,
)
@click.option(
    "-s",
    "--suite",
    help="TestRail's suite name",
    required=True,
    default=None,
    type=click.STRING,
)
@click.option(
    "-m",
    "--milestone",
    help="TestRail's milestone name",
    required=True,
    default=None,
    type=click.STRING,
)
@click.option(
    "-r",
    "--results",
    help="full path to the folder containing the JUNIT reports",
    type=click.Path(exists=True, resolve_path=True),
    required=True,
)
@click.option(
    "--tag",
    help="attribute in JUNIT report use to store requirements ids",
    required=False,
    default="VTestId",
    type=click.STRING,
)
@click.option(
    "--custom-field",
    help="TestRail's case custom field use to store the requirement id",
    required=False,
    default="custom_vteststudio_id",
    type=click.STRING,
)
@click.pass_context
def cli_upload(
    ctx,
    run_name: str,
    project: str,
    suite: str,
    milestone: str,
    results: str,
    tag: str,
    custom_field: str,
) -> None:
    """Upload all test case results on TestRail."""
    JunitReport.set_id_tag(tag)

    # get all available cases from the given project
    project_id, cases = enumerate_all_cases(
        base_url=ctx.obj["URL"],
        user=ctx.obj["USER"],
        password=ctx.obj["PASSWORD"],
        project_name=project,
        custom_field=custom_field,
    )
    # get the milestone id from the given milestone name for the run
    # creation
    milestone_id = search_for_milestone(
        base_url=ctx.obj["URL"],
        user=ctx.obj["USER"],
        password=ctx.obj["PASSWORD"],
        project_name=project,
        milestone_name=milestone,
    )
    # get the suite id from the given suite name for the run creation
    suite_id = search_for_suite(
        base_url=ctx.obj["URL"],
        user=ctx.obj["USER"],
        password=ctx.obj["PASSWORD"],
        project_name=project,
        suite_name=suite,
    )
    # extract the ids from each given JUNIT report and finally get the
    # tuple (TestRail id, req id, state)
    case_results = extract_test_results(results=results, cases=cases)
    if len(case_results) == 0:
        print("Could not match any JUnit result to a Testrail test case, exiting.")
        sys.exit(1)

    case_ids = [rail_id for rail_ids, _, _ in case_results for rail_id in rail_ids]
    # Create a brand new run on TestRail and associate the case ids
    run_id = create_new_run(
        base_url=ctx.obj["URL"],
        user=ctx.obj["USER"],
        password=ctx.obj["PASSWORD"],
        project_id=project_id,
        run_name=run_name,
        suite_id=suite_id,
        milestone_id=milestone_id,
        case_ids=case_ids,
    )
    # update the new created run with case statuses
    test_results = add_case_results(
        base_url=ctx.obj["URL"],
        user=ctx.obj["USER"],
        password=ctx.obj["PASSWORD"],
        run_id=run_id,
        results=case_results,
    )
    print_results(run_name, test_results)



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Instrument Control Auxiliary
****************************

:module: instrument_control

:synopsis: provide a simple interface to control instruments using SCPI protocol.

The functionalities provided in this package may be used directly inside ITF tests
using the corresponding auxiliary, but also using a CLI.

.. warning::
   |  This auxiliary can only be used with the cc_visa or cc_tcp_ip connector.
   |  It is not intended to be used with a proxy connector.
   |  One instrument is bound to one auxiliary even if the instrument has multiple channels.

.. currentmodule:: instrument_control

.. autosummary::

   pykiso.lib.auxiliaries.instrument_control_auxiliary.instrument_control_auxiliary
   pykiso.lib.auxiliaries.instrument_control_auxiliary.instrument_control_cli
   pykiso.lib.auxiliaries.instrument_control_auxiliary.lib_scpi_commands
   pykiso.lib.auxiliaries.instrument_control_auxiliary.lib_instruments

.. automodule:: pykiso.lib.auxiliaries.instrument_control_auxiliary.instrument_control_auxiliary
   :members:

.. automodule:: pykiso.lib.auxiliaries.instrument_control_auxiliary.instrument_control_cli
   :members:

.. automodule:: pykiso.lib.auxiliaries.instrument_control_auxiliary.lib_scpi_commands
   :members:

.. automodule:: pykiso.lib.auxiliaries.instrument_control_auxiliary.lib_instruments
   :members:
"""

from . import instrument_control_auxiliary, lib_instruments, lib_scpi_commands
from .instrument_control_auxiliary import InstrumentControlAuxiliary
from .lib_instruments import REGISTERED_INSTRUMENTS, SCPI_COMMANDS_DICT
from .lib_scpi_commands import LibSCPI



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Simulated Auxiliary
*******************

:module: simulated_auxiliary

:synopsis: auxiliary used to simulate a virtual Device Under Test(DUT)

.. currentmodule:: simulated_auxiliary
"""

import itertools
import logging

from pykiso import AuxiliaryInterface, message

from .simulation import Simulation

ACK = message.MessageType.ACK
ABORT = message.MessageCommandType.ABORT
PING = message.MessageCommandType.PING

log = logging.getLogger(__name__)


class SimulatedAuxiliary(AuxiliaryInterface):
    """Custom auxiliary use to simulate a virtual DUT."""

    def __init__(self, com=None, **kwargs):
        """Initialize attributes.

        :param com: configured channel
        """
        super().__init__(**kwargs)
        self.channel = com
        self.context = Simulation()
        self.playbook = None
        self.ts_tc = None
        log.internal_debug(f"com is {com}")

    def _create_auxiliary_instance(self) -> bool:
        """Open current associated channel.

        :return: always True
        """
        log.internal_info("Create auxiliary instance")
        log.internal_info("Enable channel")
        self.channel.open()
        return True

    def _delete_auxiliary_instance(self) -> bool:
        """Close current associated channel.

        :return: always True
        """
        log.internal_info("Delete auxiliary instance")
        self.channel.close()
        return True

    def _run_command(self, cmd_message: message.Message, cmd_data: bytes = None) -> None:
        """Not used."""
        pass

    def _abort_command(self) -> None:
        """Not used."""
        pass

    def _receive_message(self, timeout_in_s: float) -> message.Message or None:
        """Read current received message from the associated channel
        and play a pre-configured scenario depending on message
        test suite and case number.

        :param timeout_in_s: timeout applied on receive event.

        :return: Message if successful, otherwise None
        """
        responses = []

        recv_response = self.channel.cc_receive(timeout_in_s)

        msg = recv_response.get("msg")

        # don't advance simulation if no message or ACK has been received
        if msg is None or msg.msg_type == ACK:
            return msg

        # if message is PING, respond accordingly
        if msg.sub_type == PING:
            responses = next(iter(self.context.handle_ping_pong()))
            self._send_responses(responses(msg))
            return msg

        # determine pre-defined scenario if ts/tc changes
        # the ABORT command always comes with ts=0, tc=0, but we still want to
        # advance the playbook.
        if msg.sub_type != ABORT and (msg.test_suite, msg.test_case) != self.ts_tc:
            self.ts_tc = msg.test_suite, msg.test_case
            cur_scenario = self.context.get_scenario(*self.ts_tc)
            # cycle creates an iterator and makes it possible to replay the current
            # playbook.
            self.playbook = itertools.cycle(cur_scenario)

        play = next(self.playbook)
        responses = play(msg) if play is not None else []
        self._send_responses(responses)

        return msg

    def _send_responses(self, responses: message.Message) -> None:
        """Send response over CChannel.

        :param responses: TestApp protocol message type(ack, abort...)
        """
        for response in responses:
            self.channel._cc_send(msg=response)



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Configuration collection
************************

:module: global_config

:synopsis: expose the framework configuration to the user's scope

.. currentmodule:: global_config

"""
import functools
import json
import threading
from types import SimpleNamespace
from typing import Any, Callable

import click

import pykiso.cli as cli


class Singleton(type):
    """Thread safe Singleton pattern implementation."""

    #: store the unique created class instance
    _instance = None
    #: used to safely return the class instance
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs) -> Any:
        """Ensure that one and only one instance of the class is created.

        :param args: positonal arguments
        :param kwargs: named arguments

        :return: created class instance
        """
        if not cls._instance:
            with cls._lock:
                if not cls._instance:
                    cls._instance = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instance


class ProtectedNamespace(SimpleNamespace):
    """Represent a basic write protected object instance."""

    def __init__(self, **kwargs) -> None:
        """Initialize SimpleNamespace attributes.

        :param kwargs: named arguments
        """
        super().__init__(**kwargs)

    def __setattr__(self, attr: str, value: Any) -> None:
        """Protect instance attributes from writing.

        :param attr: attribute's name
        :param value: value to apply

        :raises AttributeError: when value assignation is intended
        """
        raise AttributeError(f"Attribute {attr} is not writable")


class GlobalConfig(metaclass=Singleton):
    """Container object used by the user to have access to all
    configuration information coming from the different levels of the
    framework(yaml, cli...).
    """

    def __init__(self) -> None:
        """Initialize cli and yaml attributes by taking the content of
        Grabber collector object.
        """
        self.cli = None
        self.yaml = None


class Grabber:
    """Responsible to collect configuration information at different
    levels of the framework."""

    @staticmethod
    def create_config_object(config: dict) -> ProtectedNamespace:
        """Create a configuration object by converting the given
        dictionary into a object representation.

        :param config: configuration's content

        :return: configuration's content as an object representation
        """
        str_conf = json.dumps(config)
        return json.loads(str_conf, object_hook=lambda x: ProtectedNamespace(**x))

    @staticmethod
    def grab_yaml_config(func: Callable) -> Callable:
        """Collect all parsed yaml information.

        :param func: decorated class

        :return: decorator inner function
        """

        @functools.wraps(func)
        def grab_inner(*args, **kwargs) -> dict:
            """Grab the return values from yaml configuration parser
            method and store it in the GlobalConfig instance.

            :param args: positonal arguments
            :param kwargs: named arguments

            :return: parsed yaml configuration information
            """
            dict_config = func(*args, **kwargs)
            object_config = Grabber.create_config_object(dict_config)
            GlobalConfig().yaml = object_config
            return dict_config

        return grab_inner

    @staticmethod
    def grab_cli_config(func) -> Callable:
        """Collect all cli configuration information.

        :param func: decorated class

        :return: decorator inner function
        """

        @functools.wraps(func)
        def grab_inner(click_context: click.Context, *args, **kwargs) -> None:
            """Grab the given values from cli entry point level and
            store it in the GlobalConfig instance.

            :param click_context: click context
            :param args: positonal arguments
            :param kwargs: named arguments
            """
            click_args = cli.eval_user_tags(click_context)
            # replace all dashes with underscore to make valid variable names out of the tag names
            click_args = {tag_name.replace("-", "_"): tag_value for tag_name, tag_value in click_args.items()}
            object_config = Grabber.create_config_object({**click_args, **kwargs})
            GlobalConfig().cli = object_config
            return func(click_context, *args, **kwargs)

        return grab_inner



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Virtual Communication Channel for tests
***************************************

:module: cc_example


.. warning: Still under test
"""

import logging
import threading
import time
from typing import Dict, Optional

from pykiso import connector, message

log = logging.getLogger(__name__)


class CCExample(connector.CChannel):
    """Only use for development purpose.

    This channel simply handle basic TestApp response mechanism.
    """

    def __init__(self, name: Optional[str] = None, **kwargs):
        """Initialize attributes.

        :param name: name of the communication channel
        """
        super().__init__(name=name)
        self.options = kwargs
        self.last_received_message = None
        self.report_requested_message = None
        self.lock = threading.Lock()

    def _cc_open(self) -> None:
        """Open the channel."""
        log.internal_info("open channel")

    def _cc_close(self) -> None:
        """Close the channel."""
        log.internal_info("close channel")

    def _cc_send(self, msg: bytes, **kwargs) -> None:
        """Sends the message on the channel.

        :param msg: message to send.
        :param kwargs: not used

        """
        with self.lock:
            log.internal_debug("Send: {}".format(msg))
            # Exit if ack sent
            msg = message.Message.parse_packet(msg)
            if msg.get_message_type() == message.MessageType.ACK:
                return

            # Else save received message
            self.last_received_message = msg.serialize()

            # Check if it is a run test-case message and save it for the report
            if (
                msg.sub_type in set(message.MessageCommandType)
                and msg.sub_type != message.MessageCommandType.PING
                and msg.sub_type != message.MessageCommandType.ABORT
            ):
                self.report_requested_message = msg.serialize()

    def _cc_receive(self, timeout: float = 0.1) -> Dict[str, Optional[bytes]]:
        """Craft a Message that would be received from the device under test.

        :param timeout: not use

        :return: dictionary containing the received bytes if successful, otherwise None
        """
        with self.lock:
            received_message = None
            # Simulate a real wait for message
            time.sleep(1e-3)
            if self.last_received_message is not None:
                # Transform into ack
                received_message = message.Message.parse_packet(self.last_received_message)
                received_message.msg_type = message.MessageType.ACK
                received_message.sub_type = message.MessageAckType.ACK
                # Delete the stored raw message
                received_message = received_message.serialize()
                self.last_received_message = None
                log.internal_debug("Receive: {}".format(received_message))
            elif self.report_requested_message is not None:
                # Transform message to ACK
                received_message = message.Message.parse_packet(self.report_requested_message)
                received_message.msg_type = message.MessageType.REPORT
                received_message.sub_type = message.MessageReportType.TEST_PASS
                # Delete the stored raw message
                received_message = received_message.serialize()
                self.report_requested_message = None
                log.internal_debug("Receive: {}".format(received_message))

            return {"msg": received_message}



import json
import tempfile
from pathlib import Path

import requests
import xmltodict
from junitparser.cli import merge as merge_junit_xml
from requests.auth import AuthBase

from ...tool.xray.xray_report import create_result_dictionary, reformat_xml_results

API_VERSION = "api/v2/"
AUTHENTICATE_ENDPOINT = "/api/v2/authenticate"


class XrayException(Exception):
    """Raise when sending the post request is unsuccessful."""

    def __init__(self, message=""):
        self.message = message


class ClientSecretAuth(AuthBase):
    """Bearer authentication with Client ID and a Client Secret."""

    def __init__(self, base_url: str, client_id: str, client_secret: str, verify: bool | str = True) -> None:
        if base_url.endswith("/"):
            base_url = base_url[:-1]
        self.base_url = base_url
        self.client_id = client_id
        self.client_secret = client_secret
        self.verify = verify

    @property
    def endpoint_url(self) -> str:
        """Return full URL to the authenticate server."""
        return f"{self.base_url}{AUTHENTICATE_ENDPOINT}"

    def __call__(self, r: requests.PreparedRequest) -> requests.PreparedRequest:
        headers = {"Content-type": "application/json", "Accept": "text/plain"}
        auth_data = {"client_id": self.client_id, "client_secret": self.client_secret}

        try:
            response = requests.post(self.endpoint_url, data=json.dumps(auth_data), headers=headers, verify=self.verify)
        except requests.exceptions.ConnectionError as exc:
            err_message = f"ConnectionError: cannot authenticate with {self.endpoint_url}"
            raise XrayException(err_message) from exc
        else:
            auth_token = response.text.replace('"', "")
            r.headers["Authorization"] = f"Bearer {auth_token}"
        return r


class XrayInterface:
    """Xray Interface for both REST API and GraphQL API."""

    def __init__(self, base_url: str, client_id: str, client_secret: str, verify: bool | str = True) -> None:
        self.base_url = base_url[:-1] if base_url.endswith("/") else base_url
        self.verify = verify

        # Create authentication
        self.auth = ClientSecretAuth(base_url=base_url, client_id=client_id, client_secret=client_secret, verify=verify)

        # Define the endpoints
        self.rest_endpoint = f"{self.base_url}/api/v2/import/execution"
        self.graphql_endpoint = f"{self.base_url}/api/v2/graphql"

    @property
    def endpoint_url(self) -> str:
        """Return full URL complete url to send the post request to the xray server."""
        return self.rest_endpoint

    def publish_xml_result(self, data: dict) -> dict[str, str]:
        """
        Publish the xml test results to xray.

        :param data: the test results

        :return: the content of the post request to create the execution test ticket: its id, its key, and its issue
        """
        print("Uploading test results to Xray...")
        # construct the request header
        headers = {"Accept": "application/json", "Content-Type": "application/json"}
        try:
            query_response = requests.request(
                method="POST", url=self.rest_endpoint, headers=headers, json=data, auth=self.auth, verify=self.verify
            )
        except requests.exceptions.ConnectionError:
            raise XrayException(f"Cannot connect to JIRA service at {self.rest_endpoint}")
        else:
            query_response.raise_for_status()

        return json.loads(query_response.content)

    def get_internal_id_from_key(self, jira_key: str) -> str:
        """
        Convert the Jira key to the internal issueId

        :param jira_key: the jira key of the test execution (e.g ABC-123)

        :return: the corresponding internal xray issue id
        """
        headers = {"Content-Type": "application/json", "Authorization": f"Bearer {self.auth}"}

        # GraphQL query
        query = (
            """
        query {
            getTestExecutions(jql: "key='%s'", limit: 1) {
                results {
                    issueId
                }
            }
        }
        """
            % jira_key
        )

        payload = {"query": query}

        try:
            query_response = requests.request(
                method="POST",
                url=self.graphql_endpoint,
                headers=headers,
                json=payload,
                auth=self.auth,
                verify=self.verify,
            )

            data = query_response.json()

            if (
                "data" in data
                and "getTestExecutions" in data["data"]
                and "results" in data["data"]["getTestExecutions"]
                and len(data["data"]["getTestExecutions"]["results"]) > 0
            ):
                issue_id = data["data"]["getTestExecutions"]["results"][0]["issueId"]
                return issue_id

        except requests.exceptions.ConnectionError:
            raise XrayException(f"Cannot connect to JIRA service at {self.graphql_endpoint}")
        else:
            query_response.raise_for_status()
            return None

    def get_test_execution_results(self, test_execution_id: str) -> list[str]:
        """
        Get the tests' issue ids inside the given test execution id

        :param test_execution_id: the test execution id

        :return: the list of issue ids inside the test execution results
        """

        headers = {"Content-Type": "application/json", "Authorization": f"Bearer {self.auth}"}

        # GraphQL query
        query = (
            """
        query {
            getTestExecution(issueId: "%s") {
                issueId
                tests(limit: 100) {
                    total
                    start
                    limit
                    results {
                        issueId
                        testType {
                            name
                        }
                    }
                }
            }
        }
        """
            % test_execution_id
        )

        payload = {"query": query}
        try:
            query_response = requests.request(
                method="POST",
                url=self.graphql_endpoint,
                headers=headers,
                json=payload,
                auth=self.auth,
                verify=self.verify,
            )

            data = query_response.json()

            # Extract issueIds from results
            if (
                "data" in data
                and "getTestExecution" in data["data"]
                and "tests" in data["data"]["getTestExecution"]
                and "results" in data["data"]["getTestExecution"]["tests"]
            ):
                results = data["data"]["getTestExecution"]["tests"]["results"]
                issue_ids = [result["issueId"] for result in results]
                return issue_ids

        except requests.exceptions.ConnectionError:
            raise XrayException(f"Cannot connect to JIRA service at {self.graphql_endpoint}")

        else:
            query_response.raise_for_status()

        return []

    def get_jira_keys_for_tests(self, issue_ids: list[str]) -> list[str]:
        """
        Get Jira keys for each test

        :param issue_ids: list of xray issue ids

        :return: the jira keys corresponding to the issue ids
        """

        headers = {"Content-Type": "application/json", "Authorization": f"Bearer {self.auth}"}

        jira_keys = []

        for issue_id in issue_ids:
            query = (
                """
            query {
                getTest(issueId: "%s") {
                    jira(fields: ["key", "summary"])
                    testType {
                        name
                    }
                }
            }
            """
                % issue_id
            )

            payload = {"query": query}

            try:
                query_response = requests.request(
                    method="POST",
                    url=self.graphql_endpoint,
                    headers=headers,
                    json=payload,
                    auth=self.auth,
                    verify=self.verify,
                )

                data = query_response.json()

                if (
                    "data" in data
                    and "getTest" in data["data"]
                    and data["data"]["getTest"]
                    and "jira" in data["data"]["getTest"]
                ):
                    jira_key = data["data"]["getTest"]["jira"]["key"]
                    jira_keys.append(jira_key)

            except (requests.exceptions.RequestException, json.JSONDecodeError, KeyError) as e:  # noqa F841
                continue

        return jira_keys

    def get_all_test_jira_keys(self, test_execution_id: str) -> list[str]:
        """
        Get all Jira keys for tests in a test execution

        :param test_execution_id: the test execution id (e.g ABC-1234)

        :return: the list of jira keys inside the test execution results
        """

        # Convert the provided jira key to the internal issue id
        xray_id = self.get_internal_id_from_key(test_execution_id)
        # Get all issue IDs from the test execution
        issue_ids = self.get_test_execution_results(xray_id)

        if not issue_ids:
            return []

        # Get all Jira keys for these issues
        jira_keys = self.get_jira_keys_for_tests(issue_ids)

        return jira_keys

    def upload_test_results(self, data: dict) -> dict[str, str]:
        """
        Upload all given results to xray.

        :param data: the test results

        :return: the content of the post request to create the execution test ticket: its id, its key, and its issue
        """
        return self.publish_xml_result(data=data)

    def get_jira_test_keys_from_test_execution_ticket(self, test_execution_id: str) -> list[str]:
        """
        Get the jira keys of the xray test tickets already presents in the test execution ticket.

        :param test_execution_id: the test execution id (e.g ABC-1234)

        :return: the list of jira keys inside the test execution ticket containing the test results
        """
        return self.get_all_test_jira_keys(test_execution_id)


def extract_test_results(
    path_results: Path,
    merge_xml_files: bool,
    jira_keys: list[str],
    test_execution_key: str | None = None,
    test_execution_summary: str | None = None,
    test_execution_description: str | None = None,
) -> list[str]:
    """
    Extract the test results linked to an xray test key. Filter the JUnit xml files generated by the execution of tests,
    to keep only the results of tests marked with an xray decorator. A temporary file is created with the test results.

    :param ctx: click context
    :param path_results: the path to the xml files
    :param merge_xml_files: merge all the files to return only a list with one element
    :param jira_keys: the list of jira keys inside the test execution ticket containing the test results
    :param test_execution_key: the xray's test execution ticket key where to import the test results,
        if none is specified a new test execution ticket will be created
    :param test_execution_summary: update the test execution ticket summary - otherwise, keep current summary
    :param test_execution_description: update the test execution ticket description - otherwise, keep current description

    :return: the filtered test results"""
    xml_results = []
    if path_results.is_file():
        if path_results.suffix != ".xml":
            raise RuntimeError(
                f"Expected xml file but found a {path_results.suffix} file instead, from path {path_results}"
            )
        file_to_parse = [path_results]
    elif path_results.is_dir():
        file_to_parse = list(path_results.glob("*.xml"))
        if not file_to_parse:
            raise RuntimeError(f"No xml found in following repository {path_results}")

    with tempfile.TemporaryDirectory() as xml_dir:
        if merge_xml_files and len(file_to_parse) > 1:
            xml_dir = Path(xml_dir).resolve()
            xml_path = xml_dir / "xml_merged.xml"
            merge_junit_xml(file_to_parse, xml_path, None)
            file_to_parse = [xml_path]

        # use xml to json
        for file in file_to_parse:
            with open(file) as xml_file:
                data_dict = xmltodict.parse(xml_file.read(), attr_prefix="")

            xray_dict = create_result_dictionary(
                data_dict["testsuites"]["testsuite"], jira_keys, test_execution_summary, test_execution_description
            )
            xml_results = reformat_xml_results(xray_dict, test_execution_key)
        return xml_results



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Communication Channel Via Uart
********************************

:module: cc_uart

:synopsis: Uart communication channel

.. currentmodule:: cc_uart

"""

import struct
import time
from typing import Optional

try:
    import serial
except ImportError as e:
    raise ImportError(f"{e.name} dependency missing, consider installing pykiso with 'pip install pykiso[serial]'")

from pykiso import connector, message


class IncompleteCCMsgError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class CCUart(connector.CChannel):
    """UART implementation of the coordination channel."""

    headerSize = 8
    payloadLengthIndex = 7

    WAITING_FOR_START = 0
    RECEIVING_HEADER = 1
    RECEIVING_PAYLOAD = 2
    RECEIVED_DONE = 3

    START = 0xC0
    ESC = 0xDB
    ESC_START = 0xDC
    ESC_ESC = 0xDD

    def __init__(self, serialPort, baudrate=9600, **kwargs):
        # Initialize the super class
        super().__init__(**kwargs)
        # Initialize the serial connection
        self.serial = serial.Serial(timeout=1)
        self.serial.port = serialPort
        self.serial.baudrate = baudrate
        self.serial.paritiy = serial.PARITY_NONE
        # Set a timeout to send the signal to the GIL to change thread.
        # In case of a multi-threading system, all tasks will be called one after the other.
        self.timeout = 1e-6

    def _cc_open(self):
        self.serial.open()

    def _cc_close(self):
        self.serial.close()

    def _cc_send(self, msg):
        rawPacket = msg.serialize()
        # Use CRC to verify content
        crc = self._calculate_crc32(rawPacket)
        rawPacket = struct.pack(">H", crc) + rawPacket  # Force big endian notation
        self._send_using_slip(rawPacket)

    def _cc_receive(self, timeout=0.00001):
        self.serial.timeout = timeout or self.timeout

        receivingState = self.WAITING_FOR_START
        bytesToRead = 10
        rawPacket = []

        while bytesToRead > 0:
            singleByteRead = self.serial.read(1)

            if 0 == len(singleByteRead):
                return None

            if self.WAITING_FOR_START == receivingState:
                if self.START == singleByteRead[0]:
                    receivingState = self.RECEIVING_HEADER
            else:
                if self.START == singleByteRead[0]:
                    bytesToRead = 10
                    rawPacket = []
                    receivingState = self.RECEIVING_HEADER
                    # TODO: dei9bue something went wrong
                else:
                    bytesToRead -= 1

                    if self.ESC == singleByteRead[0]:
                        singleByteRead = self.serial.read(1)

                        if self.ESC_START == singleByteRead[0]:
                            rawPacket.append(self.START)
                        elif self.ESC_ESC == singleByteRead[0]:
                            rawPacket.append(self.ESC)
                    else:
                        rawPacket.append(singleByteRead[0])

                if 0 == bytesToRead:
                    if self.RECEIVING_HEADER == receivingState:
                        bytesToRead = singleByteRead[0]
                        if bytesToRead:
                            receivingState = self.RECEIVING_PAYLOAD
                        else:
                            receivingState = self.RECEIVED_DONE
                    elif self.RECEIVING_PAYLOAD == receivingState:
                        receivingState = self.RECEIVED_DONE

            if receivingState == self.RECEIVED_DONE:
                expectedCRC = ((rawPacket[0] & 0xFF) << 8) + (rawPacket[1] & 0xFF)
                rawPacket = rawPacket[2 : len(rawPacket)]
                calculatedCRC = self._calculate_crc32(rawPacket)

                if calculatedCRC != expectedCRC:
                    return None
        # Reception was a success, we need now to convert the list into a array of bytes
        binary_message = b"".join(struct.pack("B", value) for value in rawPacket)
        return message.Message.parse_packet(binary_message)

    def _send_using_slip(self, rawPacket):
        self.serial.write(self.START.to_bytes(1, byteorder="big"))

        for aByte in rawPacket:
            if aByte == self.START:
                self.serial.write(self.ESC.to_bytes(1, byteorder="big"))
                self.serial.write(self.ESC_START.to_bytes(1, byteorder="big"))
            elif aByte == self.ESC:
                self.serial.write(self.ESC.to_bytes(1, byteorder="big"))
                self.serial.write(self.ESC_ESC.to_bytes(1, byteorder="big"))
            else:
                self.serial.write(aByte.to_bytes(1, byteorder="big"))

        time.sleep(0.01)
        self.serial.flushOutput()
        return

    def _calculate_crc32(self, buffer):
        crc = 0

        CRC_BIT_SHIFT_4 = 4
        CRC_BIT_SHIFT_5 = 5
        CRC_BIT_SHIFT_8 = 8
        CRC_BIT_SHIFT_12 = 12
        CRC_BYTE_MASK = 0xFF

        for aByte in buffer:
            crc = ((crc >> (CRC_BIT_SHIFT_8)) | (crc << (CRC_BIT_SHIFT_8))) & 0xFFFF
            crc = (crc ^ (aByte & CRC_BYTE_MASK)) & 0xFFFF
            crc = (crc ^ (crc & (CRC_BYTE_MASK)) >> (CRC_BIT_SHIFT_4)) & 0xFFFF
            crc = (crc ^ ((crc << (CRC_BIT_SHIFT_12)) & 0xFFFF)) & 0xFFFF
            crc = (crc ^ (crc & (CRC_BYTE_MASK)) << (CRC_BIT_SHIFT_5)) & 0xFFFF

        return crc & 0xFFFF



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Simulation
**********

:module: simulation

:synopsis: map virtual DUT behavior with test case/suite id

.. currentmodule:: simulation

.. warning:: Still under test
"""
from collections import defaultdict

from .response_templates import ResponseTemplates
from .scenario import Scenario, TestScenario


class Simulation:
    """Simulate a virtual DUT, by playing pre-defined scenario depending on
    test case and test suite id.
    """

    def __init__(self):
        """Initialize attributes and mapping."""

        self.map_context = defaultdict(lambda: self.handle_default_response)
        self.map_context[0, 0] = self.handle_ping_pong
        self.map_context[1, 1] = TestScenario.handle_successful

        # test case setup mapping
        setup_tc = TestScenario.VirtualTestCase.Setup
        self.map_context[1, 2] = setup_tc.handle_failed_report_setup
        self.map_context[1, 3] = setup_tc.handle_not_implemented_report_setup
        self.map_context[1, 4] = setup_tc.handle_lost_communication_during_setup_ack
        self.map_context[1, 5] = setup_tc.handle_lost_communication_during_setup_report

        # test case run mapping
        run_tc = TestScenario.VirtualTestCase.Run
        self.map_context[1, 6] = run_tc.handle_failed_report_run
        self.map_context[1, 7] = run_tc.handle_not_implemented_report_run
        self.map_context[1, 8] = run_tc.handle_lost_communication_during_run_ack
        self.map_context[1, 9] = run_tc.handle_lost_communication_during_run_report
        self.map_context[1, 14] = TestScenario.handle_successful
        self.map_context[1, 15] = run_tc.handle_successful_report_run_with_log
        self.map_context[1, 16] = run_tc.handle_failed_report_run_with_log

        # test case teardown mapping
        td_tc = TestScenario.VirtualTestCase.Teardown
        self.map_context[1, 10] = td_tc.handle_failed_report_teardown
        self.map_context[1, 11] = td_tc.handle_not_implemented_report_teardown
        self.map_context[1, 12] = td_tc.handle_lost_communication_during_teardown_ack
        self.map_context[1, 13] = td_tc.handle_lost_communication_during_teardown_report

        # test suite setup mapping
        setup_ts = TestScenario.VirtualTestSuite.Setup
        self.map_context[2, 0] = setup_ts.handle_failed_report_setup
        self.map_context[3, 0] = setup_ts.handle_not_implemented_report_setup
        self.map_context[4, 0] = setup_ts.handle_lost_communication_during_setup_ack
        self.map_context[5, 0] = setup_ts.handle_lost_communication_during_setup_report

        # test suite teardown mapping
        td_ts = TestScenario.VirtualTestSuite.Teardown
        self.map_context[6, 0] = td_ts.handle_failed_report_teardown
        self.map_context[7, 0] = td_ts.handle_not_implemented_report_teardown
        self.map_context[8, 0] = td_ts.handle_lost_communication_during_teardown_ack
        self.map_context[9, 0] = td_ts.handle_lost_communication_during_teardown_report

    def get_scenario(self, test_suite_id: int, test_case_id: int) -> Scenario:
        """Return the selected scenario mapped with the received
        test case and test suite id.

        :param test_suite_id: current test suite id
        :param test_case_id: current test case id

        :return: scenario instance containing all steps
        """
        scenario = self.map_context[test_suite_id, test_case_id]()
        return scenario

    def handle_ping_pong(self) -> Scenario:
        """Return a scenario to handle init ping pong exchange.

        :return: scenario instance containing all steps
        """
        return Scenario([ResponseTemplates.ack])

    def handle_default_response(self) -> Scenario:
        """Return a scenario to handle DUT default behavior.

        :return: scenario instance containing all steps
        """
        return Scenario([ResponseTemplates.default])



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Integration Test Framework
**************************

:module: cli

:synopsis: Entry point to the integration test framework.

.. currentmodule:: cli

"""
import logging
import os
import pprint
import sys
import threading
import time
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import click

import pykiso.profiling  # noqa: needs to be imported to trace from the beginning

from . import __version__
from .config_parser import parse_config
from .global_config import Grabber
from .logging_initializer import change_logger_class, initialize_logging
from .test_coordinator import test_execution
from .test_setup.config_registry import ConfigRegistry
from .types import PathType

UNRESOLVED_THREAD_TIMEOUT = 10


def eval_user_tags(click_context: click.Context) -> Dict[str, List[str]]:
    """Evaluate commandline args for user tags and raise exceptions for invalid
    arguments.

    :param click_context: click context
    :raises click.NoSuchOption: if key doesn't start with "--" or has an invalid
      character like "_"
    :raises click.BadOptionUsage: no value specified for user tag
    :return: user tags with values
    """
    user_tags = {}
    user_args = click_context.args.copy()
    if not user_args:
        return user_tags
    while user_args:
        try:
            key = user_args.pop(0)
            if not key.startswith("--") or "_" in key:
                correct_key = f'{"" if key.startswith("--") else "--"}{key.replace("_", "-")}'
                raise click.NoSuchOption(option_name=key, possibilities=[correct_key])

            value = user_args.pop(0)
            user_tags[key[2:]] = value.split(",")
        except IndexError:
            raise click.BadOptionUsage(
                option_name=key,
                message=f"No value specified for tag {key}",
                ctx=click_context,
            )
    return user_tags


def check_file_extension(click_context: click.Context, param: click.Parameter, paths: Tuple[str]) -> Tuple[str]:
    """Check the path strings of all given test configuration files for yaml file
    extension and raise an exception if the check fails

    :click_context: click context (click callback requirement)
    :param: click parameter (click callback requirement)
    :paths: the paths of the files to check
    :raises click.BadParameter: if one of the files does not have .yaml or .yml
        as extension
    :return: the checked paths as they were passed
    """
    for path in paths:
        if not path.endswith((".yaml", ".yml")):
            raise click.BadParameter(f"Configuration needs to be a .yaml file, but {path} was given")
    return paths


def active_threads() -> list[tuple[str, str]]:
    """Get the names and current execution frame of all active threads except the main thread.

    :return: The the thread name and the frame of the thread
    """
    frames = sys._current_frames()
    return [(thread.name, str(frames.get(thread.ident))) for thread in threading.enumerate()][1:]


def check_and_handle_unresolved_threads(log: logging.Logger, exit_code: int, timeout: int = 10) -> None:
    """Check if there are unresolved threads and handle them.

    :param log: logger
    :param exit_code: exit code to use if unresolved threads are still running after the timeout
    :param timeout: timeout in seconds to wait for unresolved threads to be terminated

    Process for unresolved threads:
    - If there are unresolved threads, log a warning and wait for a timeout.
    - If the threads are still running after the timeout, log a fatal error and force exit.
    - If the threads are properly shut down, log a warning and exit normally.
    """
    # Skip main thread and get running threads
    running_threads = active_threads()

    if len(running_threads) > 0:
        for thread_name, frame in running_threads:
            log.warning(f"Unresolved thread {thread_name} is still running. Frame: {frame}")
        log.warning(f"Wait {timeout}s for unresolved threads to be terminated.")
        time.sleep(timeout)
        if threading.active_count() > 1:
            log.fatal(
                f"Unresolved threads {', '.join([thread_name for thread_name, _ in active_threads()])} are still running after {timeout} seconds. Force pykiso to Exit."
            )
            os._exit(exit_code)
        else:
            log.warning("Unresolved threads has been properly shut down. Normal exit.")


class CommandWithOptionalFlagValues(click.Command):
    """Custom command that allows specifying flags with a value, e.g. ``pykiso -c config.yaml --junit=./reports``."""

    def parse_args(self, ctx, args):
        """Translate any flag `--junit=value` as flag `--junit` with changed flag_value=value"""
        # filter out flags from all of the command parameters
        flags = [
            flag
            for flag in self.params
            if isinstance(flag, click.Option) and flag.is_flag and not isinstance(flag.flag_value, bool)
        ]
        # iterate over all user provided arguments to match the flags with format '--flag=value'
        for arg_index, arg in enumerate(args):
            arg = arg.split("=")
            if len(arg) != 2:
                continue
            arg_name, arg_value = arg
            for flag in flags:
                # if the argument is a flag with a value, rewrite the argument as a regular flag with the appropriate value
                if arg_name in flag.opts:
                    flag.flag_value = arg_value
                    args[arg_index] = arg_name
                    break

        result_args = super(CommandWithOptionalFlagValues, self).parse_args(ctx, args)
        return result_args


@click.command(
    context_settings={
        "help_option_names": ["-h", "--help"],
        "ignore_unknown_options": True,
        "allow_extra_args": True,
    },
    cls=CommandWithOptionalFlagValues,
)
@click.option(
    "-c",
    "--test-configuration-file",
    required=True,
    type=click.Path(exists=True, dir_okay=False, readable=True),
    multiple=True,
    callback=check_file_extension,
    help="path to the test configuration file (in YAML format)",
)
@click.option(
    "-l",
    "--log-path",
    required=False,
    default=None,
    type=click.Path(writable=True),
    multiple=True,
    help="path to log-file or folder. If not set will log to STDOUT",
)
@click.option(
    "-s",
    "--log-file-strategy",
    required=False,
    default=None,
    type=click.Choice(["testRun", "testCase"], case_sensitive=True),
    help="Strategy for the log file creation, 'testRun' for one log file per run, 'testCase' for one log file per test case. Default is one log file for all tests."
    "This flag need to be used with --log-path option.",
)
@click.option(
    "--log-level",
    required=False,
    default="INFO",
    type=click.Choice(
        "DEBUG INFO WARNING ERROR".split(" "),
        case_sensitive=False,
    ),
    help="set the verbosity of the logging",
)
@click.option(
    "--junit",
    is_flag=True,
    flag_value="reports",
    help="Enable junit reports, if you want to save report in specific dir relative to current or .xml file, use --junit=(name or dir). The default dir is './reports' and default name is '%Y-%m-%d_%H-%M-%S-{config_name}.xml'",
)
@click.option(
    "--step-report",
    required=False,
    default=None,
    type=click.Path(writable=True),
    help="generate the step report at the specified path",
)
@click.option(
    "--failfast",
    is_flag=True,
    help="stop the test run on the first error or failure",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    required=False,
    help="activate the internal framework logs",
)
@click.option(
    "-p",
    "--pattern",
    type=click.STRING,
    required=False,
    help="test filter pattern, e.g. 'test_suite_1.py' or 'test_*.py'. Or even more granularly 'test_suite_1.py::TestClass::test_name'",
)
@click.option(
    "--logger",
    type=click.STRING,
    required=False,
    help="use the specified logger class in pykiso",
)
@click.version_option(__version__)
@click.pass_context
@Grabber.grab_cli_config
def main(
    click_context: click.Context,
    test_configuration_file: Tuple[PathType],
    log_path: Tuple[PathType] | None = None,
    log_file_strategy: str | None = None,
    log_level: str = "INFO",
    report_type: str = "text",
    step_report: Optional[PathType] = None,
    pattern: Optional[str] = None,
    failfast: bool = False,
    verbose: bool = False,
    logger: Optional[str] = None,
    junit: Optional[str] = None,
):
    """Embedded Integration Test Framework - CLI Entry Point.

    TAG Filters: any additional option to be passed to the test as tag through
    the pykiso call. Multiple values must be separated with a comma.

    For example: pykiso -c your_config.yaml --branch-level dev,master --variant delta

    \f
    :param click_context: click context
    :param test_configuration_file: path to the YAML config file
    :param log_path: path to existing directories or files to write logs to
    :param log_file_strategy: Strategy for the log file creation, 'testRun' for one log file per run, 'testCase'
        for one log file per test case. Default is one log file for all tests."
    :param log_level: any of DEBUG, INFO, WARNING, ERROR
    :param report_type: if "test", the standard report, if "junit", a junit report is generated
    :param variant: allow the user to execute a subset of tests based on variants
    :param branch_level: allow the user to execute a subset of tests based on branch levels
    :param step_report: file path for the step report or None
    :param pattern: overwrite the pattern from the YAML file for easier test development
    :param failfast: stop the test run on the first error or failure
    :param verbose: activate logging for the whole framework
    :param logger: class of the logger that will be used in the tests
    """
    # we are expecting one log file path or as many as the provided configuration files
    if log_path and len(log_path) not in (1, len(test_configuration_file)):
        raise click.UsageError(
            f"Mismatch: {len(log_path)} log files were provided for {len(test_configuration_file)} yaml configuration files"
        )

    if log_file_strategy is not None and log_path is None:
        raise click.UsageError("The --log-file-strategy option requires the --log-path option to be set.")

    if junit is not None:
        report_type = "junit"

    # parse provided tags (any unknown option)
    user_tags = eval_user_tags(click_context)

    if logger:
        change_logger_class(log_level, verbose, logger)

    for idx, config_file in enumerate(test_configuration_file):
        yaml_name = Path(config_file).stem

        # Setup the logging
        if log_path:
            # Put all logs in one file if only one file is provided, otherwise use a new file for each yaml
            log_file = log_path[0] if len(log_path) == 1 else log_path[idx]
        else:
            log_file = None

        log = initialize_logging(log_file, log_level, verbose, report_type, yaml_name)

        # Get YAML configuration
        cfg_dict = parse_config(config_file)
        log.debug("cfg_dict:\n%s", pprint.pformat(cfg_dict))

        # Run tests
        with ConfigRegistry.provide_auxiliaries(cfg_dict):
            exit_code = test_execution.execute(
                cfg_dict, report_type, yaml_name, user_tags, step_report, pattern, failfast, junit, log_file_strategy
            )

        for handler in logging.getLogger().handlers:
            if isinstance(handler, logging.FileHandler):
                logging.getLogger().removeHandler(handler)

        check_and_handle_unresolved_threads(log, exit_code=exit_code, timeout=UNRESOLVED_THREAD_TIMEOUT)

    sys.exit(exit_code)



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Ykush Auxiliary
***************

:module: ykush_auxiliary

:synopsis: Auxiliary that can power on and off ports on an Ykush device.

.. currentmodule:: ykush_auxiliary

"""
import logging
from contextlib import contextmanager
from enum import IntEnum
from typing import Any, List, Optional, Tuple

import hid

from pykiso.auxiliary import AuxiliaryInterface

log = logging.getLogger(__name__)

# YKUSH device USB VID
YKUSH_USB_VID = 0x04D8
# YKUSH PIDs when in normal operation mode: YKUSH beta, YKUSH, YKUSH3, YKUSHXS
YKUSH_USB_PID_LIST = (0x0042, 0xF2F7, 0xF11B, 0xF0CD)

# YKUSH device USB comm declarations
YKUSH_USB_TIMEOUT = 1000  # timeout in ms
YKUSH_USB_PACKET_SIZE = 64
YKUSH_USB_PACKET_PAYLOAD_SIZE = 20

# YKUSH device protocol status declarations
YKUSH_PROTO_OK_STATUS = 1

# YKUSH port state meaning declarations
YKUSH_PORT_STATE_ON = 1
YKUSH_PORT_STATE_OFF = 0
YKUSH_PORT_STATE_ERROR = 255
YKUSH_PORT_STATE_DICT = {0: "OFF", 1: "ON", 255: "ERROR"}


class YkushError(Exception):
    """General Ykush specific exception used as basis for all others."""

    pass


class YkushDeviceNotFound(YkushError):
    """Raised when no Ykush device is found."""

    pass


class YkushStatePortNotRetrieved(YkushError):
    """Raised when the state of a port can't be retrieved."""

    pass


class YkushSetStateError(YkushError):
    """Raised when a port couldn't be switched on or off."""

    pass


class YkushPortNumberError(YkushError):
    """Raised when the port number doesn't exist."""

    pass


class PortState(IntEnum):
    OFF = 0
    ON = 1


class YkushAuxiliary(AuxiliaryInterface):
    """Auxiliary used to power on and off the ports."""

    def __init__(self, serial_number: str = None, **kwargs):
        """Initialize attribute

        :param serial: Serial number of the device to connect, if he is not defined
            then it will connect to the first Ykush device it find, defaults to None.
        :raises YkushDeviceNotFound: If no device is found or the serial number
            is not the serial of one device.
        """
        super().__init__(
            is_proxy_capable=False,
            tx_task_on=False,
            rx_task_on=False,
            connector_required=False,
            **kwargs,
        )
        self._ykush_device = None
        self._product_id = None
        self._path = None
        self.connect_device(serial_number)
        self.number_of_port = self.get_number_of_port()

    def _create_auxiliary_instance(self) -> bool:
        """Power on all port at the start.
        :return: True if successful
        """
        log.internal_info("Create auxiliary instance")
        self.set_all_ports_on()
        return True

    def _delete_auxiliary_instance(self) -> bool:
        """Power on all port to restore the ports.
        :return: always True
        """
        self.set_all_ports_on()
        log.internal_info("Auxiliary instance deleted")
        return True

    def connect_device(self, serial: int = None):
        """Find an Ykush device, will automatically connect to the first one
        it find, if you have multiple connected you have to precise the serial
        number of the device.

        :param serial: serial number of the device, defaults to None
        :raises YkushDeviceNotFound: if no ykush device is found
        """
        list_ykush_device = []
        self._path = None
        # try to locate a device
        for device in hid.enumerate(0, 0):
            if device["vendor_id"] == YKUSH_USB_VID and device["product_id"] in YKUSH_USB_PID_LIST:
                list_ykush_device.append(device["serial_number"])
                if serial is None or serial == device["serial_number"]:
                    self._product_id = device["product_id"]
                    self._path = device["path"]

        if self._path is not None:
            self._ykush_device = hid.device()
            self._ykush_device.open_path(self._path)
        else:
            if list_ykush_device == []:
                raise YkushDeviceNotFound("Could not connect to a ykush hub, no device was found.")
            else:
                raise YkushDeviceNotFound(
                    f"The serial numbers available are : {list_ykush_device}\n"
                    f"No device was found with the serial number {serial}\n"
                    if serial
                    else "",
                )

    @contextmanager
    def _open_and_close_device(self):
        """Context manager to open and close device every time we send a message
        otherwise  we will get an empty message every time in response.
        """
        self._ykush_device = hid.device()
        self._ykush_device.open_path(self._path)
        try:
            yield
        finally:
            self._ykush_device.close()

    def check_port_number(self, port_number: int):
        """Check if the port indicated is a port of the device

        :raises YkushPortNumberError: Raise error if no port has this number
        """
        if port_number not in range(1, self.number_of_port + 1):
            raise YkushPortNumberError(
                f"The port number {port_number} is not valid for the device,"
                f" it has only {self.number_of_port} ports"
            )

    @staticmethod
    def get_str_state(state: int) -> str:
        """Return the str of a state

        :param state: 1 (port on), 0 (port off)
        :return: ON, OFF
        """
        return YKUSH_PORT_STATE_DICT[state]

    def get_serial_number_string(self) -> str:
        """Returns the device serial number string"""
        with self._open_and_close_device():
            return self._ykush_device.get_serial_number_string()

    def get_number_of_port(self) -> int:
        """Returns the number of port on the ykush device"""
        # original YKUSH 1,3 port count
        count = 3
        # YKUSHXS Make it have 1 port
        if self._product_id == 0xF0CD:
            count = 1

        return count

    def get_port_state(self, port_number: int) -> PortState:
        """Returns a specific port state.

        :raises YkushStatePortNotRetrieved: If the state couldn't be retrieved
        :return: 0 if the port is off, 1 if the port is on
        """
        self.check_port_number(port_number)
        return self.get_all_ports_state()[port_number - 1]

    def get_all_ports_state(self) -> List[PortState]:
        """Returns the state of all the ports.

        :raises YkushStatePortNotRetrieved: The states couldn't be retrieved
        :return: list with 0 if a port is off, 1 if on
        """
        if self.get_firmware_version()[0] >= 1:
            recvbytes = self._raw_sendreceive([0x2A])[: self.number_of_port + 1]
            if recvbytes[0] == YKUSH_PROTO_OK_STATUS:
                return [PortState.ON if p > 0x10 else PortState.OFF for p in recvbytes[1:]]
            else:
                raise YkushStatePortNotRetrieved("The states of the ports couldn't be retrieved")
        else:
            # firmware glitch workaround
            list_state = []
            for port_number in range(1, self.number_of_port + 1):
                status, port_state = self._raw_sendreceive([0x20 | port_number])[:2]
                if status == YKUSH_PROTO_OK_STATUS:
                    list_state.append(PortState.ON if port_state > 0x10 else PortState.OFF)
                else:
                    raise YkushStatePortNotRetrieved(f"The state of the port {port_number} couldn't be retrieved")
            return list_state

    def get_firmware_version(self) -> Tuple[int, int]:
        """Returns a tuple with YKUSH firmware version in format (major, minor)."""

        status, major, minor = self._raw_sendreceive([0xF0])[:3]
        if status == YKUSH_PROTO_OK_STATUS:
            self._firmware_major_version, self._firmware_minor_version = (
                major,
                minor,
            )
        else:
            # early devices will not recognize it, figure it out from serial
            self._firmware_major_version = 1
            self._firmware_minor_version = (
                2
                if "YK2" in self.get_serial_number_string()
                else 255
                if "YKD2" in self.get_serial_number_string()
                else 0
            )
        return self._firmware_major_version, self._firmware_minor_version

    def set_port_state(self, port_number: int, state: int):
        """Set a specific port On or Off.

        :raises YkushSetStateError: if the operation had an error
        :param state: 1 (port on), 0 (port off)
        """
        str_state = self.get_str_state(state)
        self.check_port_number(port_number)
        self._raw_sendreceive([(0x10 if state == YKUSH_PORT_STATE_ON else 0x0) | port_number])

        try:
            state_port = self.get_port_state(port_number)
        except YkushStatePortNotRetrieved as e:
            raise YkushSetStateError(
                f"The state of the action to power {str_state}"
                "couldn't be confirmed because the state of the port can't be retrieved"
            ) from e
        if state_port != PortState[str_state]:
            raise YkushSetStateError(
                f"There was an error trying to power {str_state.lower()} the port,"
                f"the port {port_number} is {state_port}"
            )

    def set_port_on(self, port_number: int):
        """Power on a specific port.

        :raises YkushSetStateError: if the operation had an error
        """
        self.set_port_state(port_number, state=YKUSH_PORT_STATE_ON)

    def set_port_off(self, port_number: int):
        """Power off a specific port.

        :raises YkushSetStateError: if the operation had an error
        """
        self.set_port_state(port_number, state=YKUSH_PORT_STATE_OFF)

    def set_all_ports(self, state: int):
        """Power off or on all YKUSH ports.

        :param state: state wanted 1 for On, 0 for Off
        :raises YkushSetStateError: if the operation had an error
        """
        str_state = self.get_str_state(state)
        self._raw_sendreceive([(0x1A if state == YKUSH_PORT_STATE_ON else 0x0A)])

        try:
            states_port = self.get_all_ports_state()
        except YkushStatePortNotRetrieved as e:
            raise YkushSetStateError(
                f"The state of the action to power {str_state}"
                "couldn't be confirmed because the state of the ports can't be retrieved"
            ) from e

        if not (PortState[str_state] == states_port[0] and len(set(states_port)) == 1):
            raise YkushSetStateError(
                "There was an error during the power "
                f"{str_state.lower()},"
                f"the ports have the following states : {states_port}"
            )

    def set_all_ports_on(self):
        """Power on all YKUSH downstreams ports.

        :raises YkushSetStateError: if the operation had an error
        """
        self.set_all_ports(state=YKUSH_PORT_STATE_ON)

    def set_all_ports_off(self):
        """Power off all YKUSH downstreams ports.

        :raises YkushSetStateError: if the operation had an error
        """
        self.set_all_ports(state=YKUSH_PORT_STATE_OFF)

    def is_port_on(self, port_number: int) -> bool:
        """Check if a port is on.

        :return: True if a port is on, else False
        """
        return bool(self.get_port_state(port_number))

    def is_port_off(self, port_number: int) -> bool:
        """Check if a port is off.

        :return: True if a port is off, else False
        """
        return not bool(self.get_port_state(port_number))

    def _raw_sendreceive(self, packetarray) -> List[int]:
        """Send a message to the device and get the returned message.

        :param packetarray: packet to send to do an operation
        :return: response to the message send
        """
        with self._open_and_close_device():
            packetarray = packetarray * 2 + [0x00] * (YKUSH_USB_PACKET_SIZE - len(packetarray))
            self._ykush_device.write(packetarray)
            recvpacket = self._ykush_device.read(
                max_length=YKUSH_USB_PACKET_SIZE + 1,
                timeout_ms=YKUSH_USB_TIMEOUT,
            )

            # if not None return the bytes we actually need
            if recvpacket is None or len(recvpacket) < YKUSH_USB_PACKET_PAYLOAD_SIZE:
                return [0xFF] * YKUSH_USB_PACKET_PAYLOAD_SIZE
            return recvpacket[:YKUSH_USB_PACKET_PAYLOAD_SIZE]

    def _run_command(self, cmd_message: Any, cmd_data: Optional[bytes]) -> None:
        """Not used.

        Simply respect the interface.
        """

    def _receive_message(self, timeout_in_s: float) -> None:
        """Not used.

        Simply respect the interface.
        """



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Communication Channel Via Usb
********************************

:module: cc_usb

:synopsis: Usb communication channel

.. currentmodule:: cc_usb


.. warning: Still under test
"""

try:
    import serial
except ImportError as e:
    raise ImportError(f"{e.name} dependency missing, consider installing pykiso with 'pip install pykiso[serial]'")

from pykiso.lib.connectors import cc_uart


class CCUsb(cc_uart.CCUart):
    def __init__(self, serial_port):
        super().__init__(serial_port, baudrate=9600)

    def _cc_send(self, msg):
        raw_packet = msg.serialize()
        crc = self._calculate_crc32(raw_packet)

        raw_packet.insert(0, ((crc >> 8) & 0xFF))
        raw_packet.insert(1, (crc & 0xFF))
        self._send_using_slip(raw_packet)

    def _send_using_slip(self, raw_packet):
        slip_raw_packet = []
        slip_raw_packet.append(self.START)

        for a_byte in raw_packet:
            if a_byte == self.START:
                slip_raw_packet.append(self.ESC)
                slip_raw_packet.append(self.ESC_START)
            elif a_byte == self.ESC:
                slip_raw_packet.append(self.ESC)
                slip_raw_packet.append(self.ESC_ESC)
            else:
                slip_raw_packet.append(a_byte)
        self.serial.write(bytearray(slip_raw_packet))
        self.serial.flushOutput()
        return

    #################### todo TO DELETE IF NOT NEEDED ANYMORE #######################

    def CCwaitAfterReboot(self, retry_count, reboot_time):
        import time

        self.CCclose()
        for i in range(retry_count):
            try:
                time.sleep(reboot_time)
                self.CCopen()
                break
            except (
                serial.SerialException,
                serial.serialutil.SerialTimeoutException,
            ):
                self.logger.debug(
                    "Unable to connect to USB port after reboot." + "Retrying after " + reboot_time + " seconds"
                )



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Test configuration analysis
***************************

:module: show_tag

:synopsis: Passively load the test suites of YAML configuration files
    to show their test tags. Meant to be invoked as ``pykiso-tags``
    CLI utility.

.. currentmodule:: show_tag
"""

import csv
import json
import logging
import sys
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union
from unittest import mock  # used to disable the test auxiliaries run

import click
from tabulate import tabulate

from pykiso.config_parser import parse_config
from pykiso.exceptions import TestCollectionError
from pykiso.test_coordinator import test_case, test_execution
from pykiso.types import PathType


def get_yaml_files(config_path: PathType, recursive: bool) -> List[Path]:
    """List the YAML files passed to the CLI.

    :param config: YAML file or folder containing YAML files.
    :param recursive: boolean allowing to recurse all subfolders
        if a folder was provided.
    :raises FileNotFoundError: if a folder was provided but no
        YAML file was found.
    :raises ValueError: if a file was provided that does not have
        the .yaml extension.
    :return: list all all YAML file paths.
    """
    config = Path(config_path)
    pattern = "**/*.yaml" if recursive else "*.yaml"
    if config.is_dir():
        # search for all yaml files contained in it
        config_files = [config_file_path.resolve() for config_file_path in Path(config).glob(pattern)]
        if not config_files:
            raise FileNotFoundError(f"No YAML configuration files found at {config}")
    else:
        if not config.suffix == ".yaml":
            raise ValueError("Provided file is not in YAML format")
        config_files = [config.resolve()]

    return config_files


def get_test_cases(cfg_dict: Dict[str, List[dict]]) -> List[test_case.BasicTest]:
    """Return the list of tests meant to be run by the provided
    test configuration file.

    :param cfg_dict: loaded configuration yaml file.
    :raises ValueError: if no test suite is specified in the
        configuration file.
    :return: list of tests meant to be run by the yaml file.
    """
    if "test_suite_list" not in cfg_dict:
        raise ValueError("Provided YAML file does not define a test suite")

    test_suites = test_execution.collect_test_suites(cfg_dict["test_suite_list"])
    test_cases = [tc for ts in test_suites if ts is not None for tc in ts._tests]
    return test_cases


def get_test_tags(test_case_list: List[test_case.BasicTest]) -> Dict[str, List[str]]:
    """Return the list of tag and values contained in the test case list

    :param test_case_list: list of loaded test cases.
    :return: dictionary linking the tag names to their values.
    """
    tag_dict = {}
    # search the tag for each test case
    for testCase in test_case_list:
        if testCase.tag is None:
            continue
        for tag_name, tag_values in testCase.tag.items():
            if tag_name not in tag_dict:
                tag_dict[tag_name] = list()
            # tag values are lists of strings
            tag_dict[tag_name].extend(tag_values)
            # remove duplicated tag values
            tag_dict[tag_name] = sorted(set(tag_dict[tag_name]))

    return tag_dict


def build_result_dict(
    config_file_name: str,
    test_case_list: List[test_case.BasicTest],
    test_tags: Dict[str, list],
    show_test_cases: bool = False,
) -> Dict[str, Union[str, int]]:
    """Build a dictionary containing the test information of a single
    configuration file.

    :param config_file_name: name of the configuration file.
    :param test_case_list: list of test cases meant to be loaded by
        pykiso with the provided configuration file.
    :param test_tags: dictionary containing the tags found in the
        provided configuration file's test suites.
    :param show_test_cases: set this flag to list all of the loaded
        test cases in the result dictionary (default is False).
    :return: the result
    """
    config_result_dict = {
        "File name": config_file_name,
        "Number of tests": len(test_case_list),
    }
    if show_test_cases:
        config_result_dict.update(
            {"Test cases": "\n".join(f"{test.__module__}.{test.__class__.__qualname__}" for test in test_case_list)}
        )
    if test_tags:
        tag_dict = {tag_name: "\n".join(tag_list) for tag_name, tag_list in test_tags.items()}
        config_result_dict = {**config_result_dict, **tag_dict}
    return config_result_dict


def tabulate_test_information(
    all_tests_info: List[Dict[str, Union[str, int]]], fill_char="-"
) -> Tuple[List[str], List[list]]:
    """Format a list of test information dictionaries to a tuple
    containing a table header and data in order to be tabulated.

    :param all_tests_info: list of test information corresponding
        to one loaded configuration file.
    :param fill_char: fill character for test suites that do not
        specify any tag, defaults to "-",
    :return: the formatted table header and table value as a tuple.
    """
    # get the entry with the most tags to use it as header for the table
    table_header = list(max(res.keys() for res in all_tests_info))
    table_rows = list()
    for test_info in all_tests_info:
        row = list()
        for idx, (corresponding_header, value) in enumerate(test_info.items()):
            # ensure that the current entry's tag corresponds to the header tags
            if idx == table_header.index(corresponding_header):
                row.append(value)
            else:
                row.append(fill_char)
        # fill the rest of the row values if no tag was found
        row.extend([fill_char] * (len(table_header) - len(row)))
        table_rows.append(row)
    return table_header, table_rows


@click.command(context_settings={"help_option_names": ["-h", "--help"]})
@click.option(
    "-c",
    "--test-configuration",
    required=True,
    type=click.Path(exists=True, dir_okay=True, readable=True),
    multiple=True,
    help="Path to the test configuration file (in YAML format) or folder containing test configuration files.",
)
@click.option(
    "-o",
    "--output",
    type=click.Path(dir_okay=False),
    help="Path to the exported test analysis file. Possible formats are csv, json or txt",
)
@click.option(
    "-r",
    "--recurse-dir",
    is_flag=True,
    default=False,
    type=click.BOOL,
    help="""
    If a folder is provided, set this flag to recurse all subfolders to find
    all configuration files (in YAML format).
    """,
)
@click.option(
    "--show-tests",
    is_flag=True,
    default=False,
    type=click.BOOL,
    help="""
    Extend the resulting table with all loaded test cases per configuration file.
    """,
)
def main(
    test_configuration: Tuple[PathType],
    output: Optional[PathType] = None,
    recurse_dir: bool = False,
    show_tests: bool = False,
):
    """Embedded Integration Test Framework - Test tag analysis.

    This tool loads all test cases from each test suite listed in
    the provided configuration file (in YAML format) in order to
    extract their tags.

    The result are shown as a table with the following format:

    \b
    ╒══════════════════════════════════════╤═══════════════════╤═══════════╤════════════════╕
    │ File name                            │   Number of tests │ variant   │ branch_level   │
    ╞══════════════════════════════════════╪═══════════════════╪═══════════╪════════════════╡
    │ my_config.yaml                       │                42 │ variant1  │ daily          │
    │                                      │                   │ variant3  │                │
    ├──────────────────────────────────────┼───────────────────┼───────────┼────────────────┤

    \f
    :param test_configuration_file: path to the YAML config file
        or folder containing configuration files.
    :param output: optional path to a file to dump the tag table.
        Supported formats are csv, json and txt.
    :param recurse_dir: if a folder is provided, recurse all
        subfolders to find configuration files (default: False).
    """
    # disable logging
    logging.getLogger().setLevel(logging.CRITICAL)
    # disable running the test auxiliaries when imported by the loader
    sys.modules["pykiso.auxiliaries"] = mock.MagicMock()

    all_results: List[Dict[str, Any]] = list()

    click.echo("\nStart analyzing provided configuration file...")
    # handle multiple files or folders provided by the user
    for config in test_configuration:
        # get all YAML files passed to the CLI
        yaml_files = get_yaml_files(config, recurse_dir)

        for config_file in yaml_files:
            try:
                # parse YAML configuration file (env vars, paths, etc.)
                cfg_dict = parse_config(config_file)
            except ValueError as e:
                click.echo(f"Failed to parse config file {config_file.name}: {e}")
                continue

            try:
                # get all test cases meant to be loaded by the config file
                test_cases = get_test_cases(cfg_dict)
            except ValueError as e:
                click.echo(f"Failed to load test cases from config file {config_file.name}: {e.args[0]}")
                continue
            except TestCollectionError as e:
                click.echo(f"Failed to load test cases from config file {config_file.name}: {e}")
                continue

            tags = get_test_tags(test_cases)

            single_config_result = build_result_dict(config_file.name, test_cases, tags, show_test_cases=show_tests)
            all_results.append(single_config_result)

    table_header, table_data = tabulate_test_information(all_results)
    table = tabulate(table_data, headers=table_header, tablefmt="fancy_grid")

    click.echo("\nAll valid configuration files have been processed successfully:")
    click.echo("\n{}\n".format(table))

    if output is None:
        return

    output = Path(output)

    with open(output, "w", encoding="utf-8") as f:
        if ".json" in output.suffixes:
            json.dump(all_results, f, ensure_ascii=False, indent=2)
        elif output.suffix == ".csv":
            csv.writer(output).writerows(table_header + table_data)
        elif output.suffix == ".txt":
            f.write(table)
        else:
            click.echo(f"Unsupported export format {output.suffix}")
            return

    click.echo(f"Test configuration dumped to '{output}'")



##########################################################################
# Copyright (c) 2010-2023 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Multi test result
*****************

:module: multi_result

:synopsis: a test result proxy that allows to use a ``TestRunner`` with multiple
    `` TestResult`` subclasses

.. currentmodule:: multi_result
"""
from __future__ import annotations

import logging
import unittest
from datetime import datetime
from inspect import signature
from pathlib import Path
from typing import Any, List, Literal, Optional, Union
from unittest import TestResult, util
from unittest.case import _SubTest

from pykiso.types import ExcInfoType

from ..logging_initializer import get_internal_level, get_logging_options
from ..test_coordinator.test_case import BasicTest
from ..test_coordinator.test_suite import BaseTestSuite
from .text_result import BannerTestResult
from .xml_result import XmlTestResult

test_runner_instance: Optional[unittest.TextTestRunner] = None


class MultiTestResult:
    """Class that can take multiple test result classes and ran the
    test for all the classes.
    """

    def __init__(
        self,
        *result_classes: TestResult,
        log_file_strategy: Literal["testRun,testCase"] | None = None,
    ):
        """Initialize parameter

        :param result_classes: test result classes
        :param log_file_strategy: strategy to use for the log file
            - testRun: log file will be created for each test run
            - testCase: log file will be created for each test case
        """
        self.result_classes: list[TestResult] = result_classes
        self._log_file_strategy = log_file_strategy
        self._list_test_results: list[BasicTest] = []
        self.current_log_file: Path | None = None

    def __call__(self, *args, **kwargs) -> MultiTestResult:
        """Initialize the result classes with the parameters passed in arguments."""
        self.result_classes = [
            result(*self._get_arguments_result_class(result, *args, **kwargs)) for result in self.result_classes
        ]
        return self

    def _get_arguments_result_class(self, result_class, *args, **kwargs) -> List[Any]:
        """Function to get the argument for the result class initialisation.

        :param result_class: test result class
        :return: list of arguments to initialise the class
        """
        list_arguments = []
        args = list(args)
        sign_class = signature(result_class)
        for name in sign_class.parameters.keys():
            arg = kwargs.get(name) or (args.pop(0) if args else sign_class.parameters[name].default)
            list_arguments.append(arg)

        return list_arguments

    def __getattr__(self, name: str) -> Any:
        """Function implemented to get the attributes such as error, failures,
            successes,expectedFailures.

        :param name: name of the attribute
        """
        return getattr(self.result_classes[0], name)

    def __setattr__(self, name: str, value: Any) -> Any:
        """Function implemented to set the attributes such as failfast
            of all of the result classes

        :param name: name of the attribute
        """
        super().__setattr__(name, value)
        # condition to avoid infinite loop with the __getattr__
        if name not in ["result_classes", "_log_file_strategy", "_list_test_results", "current_log_file"]:
            for result in self.result_classes:
                setattr(result, name, value)

    @property
    def error_occurred(self) -> Optional[bool]:
        """Return if an error occurred for a BannerTestResult"""
        for result in self.result_classes:
            if hasattr(result, "error_occurred"):
                return result.error_occurred

    def startTest(self, test: Union[BasicTest, BaseTestSuite]) -> None:
        """Call the startTest function for all result classes.

        :param test: running testcase
        """
        self.handle_log_file_strategy(test)

        for result in self.result_classes:
            result.startTest(test)

    def handle_log_file_strategy(self, test: Union[BasicTest, BaseTestSuite]) -> None:
        """Handle the log file strategy for the given test.

        :param test: The test case or test suite being executed.
        """
        if self._log_file_strategy is None:
            return

        log_options = get_logging_options()

        if test.__class__ not in self._list_test_results or self._log_file_strategy == "testRun":
            log_file_name = util.strclass(test.__class__).replace(".", "_").replace("-", "_")
            if self._log_file_strategy == "testRun":
                log_file_name += "_" + test._testMethodName

            log_file_name += f"_{datetime.today().strftime('%Y%d%m%H%M%S')}.log"
            self.current_log_file = log_options.log_path.parent / log_file_name

        # Setup the file handler for logging
        log_format = logging.Formatter("%(asctime)s [%(levelname)s] %(module)s:%(lineno)d: %(message)s")
        root_logger = logging.getLogger()
        file_handler = logging.FileHandler(self.current_log_file, "a+")
        file_handler.setFormatter(log_format)
        file_handler.name = "strategy_log_file_handler"
        # always include internal logs in log files
        file_handler.setLevel(get_internal_level(log_options.log_level))
        root_logger.addHandler(file_handler)

        # Add the log file to the test runner instance to get the banner information
        test_runner_instance.stream.stream.multifile_handler.add_file(self.current_log_file)
        self._list_test_results.append(test.__class__)

    def startTestRun(self) -> None:
        """Call the startTestRun function for all result classes."""
        for result in self.result_classes:
            result.startTestRun()

    def stopTestRun(self) -> None:
        """Call the stopTestRun function for all result classes."""
        for result in self.result_classes:
            result.stopTestRun()

    def stop(self) -> None:
        """Call the stop function for all result classes."""
        for result in self.result_classes:
            result.stop()

    def stopTest(self, test: Union[BasicTest, BaseTestSuite]) -> None:
        """Call the stopTest function for all result classes.

        :param test: running testcase
        """
        for result in self.result_classes:
            result.stopTest(test)

        if self._log_file_strategy is None:
            return
        # Remove the log file
        test_runner_instance.stream.stream.multifile_handler.remove_file(self.current_log_file)
        root_logger = logging.getLogger()
        for handler in root_logger.handlers:
            if handler.name == "strategy_log_file_handler":
                root_logger.removeHandler(handler)
                handler.close()

    def addSuccess(self, test: Union[BasicTest, BaseTestSuite]) -> None:
        """Call the addSuccess function for all result classes.

        :param test: running testcase
        """
        for result in self.result_classes:
            result.addSuccess(test)

    def addSkip(self, test: Union[BasicTest, BaseTestSuite], reason: str) -> None:
        """Call the addSkip function for all result classes.

        :param test: running testcase
        :param reason: reason to skip the test
        """
        for result in self.result_classes:
            result.addSkip(test, reason)

    def addUnexpectedSuccess(self, test: Union[BasicTest, BaseTestSuite]) -> None:
        """Call the addUnexpectedSuccess function for all result classes.

        :param test: running testcase
        """
        for result in self.result_classes:
            result.addUnexpectedSuccess(test)

    def addExpectedFailure(
        self,
        test: Union[BasicTest, BaseTestSuite],
        err: ExcInfoType,
    ) -> None:
        """Call the addExpectedFailure function for all result classes.

        :param test: running testcase
        :param err: tuple returned by sys.exc_info
        """
        for result in self.result_classes:
            result.addExpectedFailure(test, err)

    def addSubTest(
        self,
        test: Union[BasicTest, BaseTestSuite],
        subtest: _SubTest,
        err: ExcInfoType,
    ) -> None:
        """Call the addSubTest function for all result classes.

        :param test: running testcase
        :param subtest: subtest run
        :param err: tuple returned by sys.exc_info
        """
        for result in self.result_classes:
            result.addSubTest(test, subtest, err)

    def addError(
        self,
        test: Union[BasicTest, BaseTestSuite],
        err: ExcInfoType,
    ) -> None:
        """Call the addError function for all result classes.

        :param test: running testcase
        :param err: tuple returned by sys.exc_info
        """
        for result in self.result_classes:
            result.addError(test, err)

    def addFailure(
        self,
        test: Union[BasicTest, BaseTestSuite],
        err: ExcInfoType,
    ) -> None:
        """Call the addFailure function for all result classes.

        :param test: running testcase
        :param err: tuple returned by sys.exc_info
        """
        for result in self.result_classes:
            result.addFailure(test, err)

    def printErrorList(self, flavour: str, errors: List[tuple]) -> None:
        """Call printErrorList function once, it will first try to call the
        function for a BannerTestResult first but if none are defined it will
        call the function of the first result class.

        :param flavour: failure reason
        :param errors: list of failed tests with their error message
        """
        for result in self.result_classes:
            if isinstance(result, BannerTestResult):
                return result.printErrorList(flavour, errors)
        return self.result_classes[0].printErrorList(flavour, errors)

    def printErrors(self) -> None:
        """Call printErrors function once, it will first try to call the function
        for a BannerTestResult first but if none are defined it will call the
        function of the first result class.
        """
        for result in self.result_classes:
            if isinstance(result, BannerTestResult):
                return result.printErrors()
        return self.result_classes[0].printErrors()

    def getDescription(self, test: Union[BasicTest, BaseTestSuite]) -> str:
        """Call getDescription function once, it will first try to call the
        function for a BannerTestResult first but if none are defined it will
        call the function of the first result class available.

        :param test: running testcase
        """
        for result in self.result_classes:
            if isinstance(result, BannerTestResult):
                return result.getDescription(test)
        return self.result_classes[0].getDescription(test)

    def generate_reports(self, test_runner) -> None:
        """Call the generate_report function for all classes in which the
        function is defined.

        :param test_runner: runner class of the test
        """
        for result in self.result_classes:
            if hasattr(result, "generate_reports"):
                result.generate_reports(test_runner)



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Double Threaded based Auxiliary Interface
*****************************************

:module: dt_auxiliary

:synopsis: common double threaded based auxiliary interface

.. currentmodule:: dt_auxiliary

"""
import abc
import enum
import functools
import logging
import queue
import threading
import warnings
from typing import Any, Callable, List, Optional

from ..auxiliary import AuxCommand, AuxiliaryInterface, close_connector, flash_target, open_connector
from ..exceptions import AuxiliaryCreationError, AuxiliaryNotStarted
from ..logging_initializer import add_internal_log_levels, initialize_loggers

log = logging.getLogger(__name__)


class DTAuxiliaryInterface(AuxiliaryInterface):
    def __init__(
        self,
        name: str = None,
        is_proxy_capable: bool = False,
        connector_required: bool = True,
        activate_log: List[str] = None,
        tx_task_on=True,
        rx_task_on=True,
        auto_start: bool = True,
    ) -> None:
        super().__init__(
            name,
            is_proxy_capable,
            connector_required,
            activate_log,
            tx_task_on,
            rx_task_on,
            auto_start,
        )
        warnings.warn(
            "The DTAuxiliaryInterface is deprecated and has been renamed to AuxiliaryInterface. Please use AuxiliaryInterface instead.",
            category=FutureWarning,
        )
